# -*- coding: utf-8 -*-
#
# Copyright (c) 2015, Alcatel-Lucent Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its contributors
#       may be used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import json

from .nurest_connection import HTTP_METHOD_PUT, HTTP_METHOD_GET
from .nurest_request import NURESTRequest
from .nurest_object import NURESTObject
from .nurest_session import _NURESTSessionCurrentContext

from .utils import Sha1


class NURESTBasicUser(NURESTObject):
    """ NURESTBasicUser defines a user that can log in.

        Only one NURESTBasicUser can be connected at a time.
    """

    __default_user = None

    def __init__(self):
        """ Initializes user """

        super(NURESTBasicUser, self).__init__()

        self._api_url = None
        self._new_password = None

        self._user_name = None
        self._password = None
        self._api_key = None

        self.expose_attribute(local_name='user_name', remote_name='userName', attribute_type=str)
        self.expose_attribute(local_name='password', attribute_type=str)
        self.expose_attribute(local_name='api_key', remote_name='APIKey', attribute_type=str)

    # Properties

    @property
    def user_name(self):
        """ Get user_name """

        return self._user_name

    @user_name.setter
    def user_name(self, user_name):
        """ Set user_name """

        self._user_name = user_name

    @property
    def password(self):
        """ Get password """

        return self._password

    @password.setter
    def password(self, password):
        """ Set password """

        self._password = password

    @property
    def api_key(self):
        """ Get API Key """

        return self._api_key

    @api_key.setter
    def api_key(self, api_key):
        """ Set API Key """

        self._api_key = api_key

    # Class Methods

    @classmethod
    def get_default_user(cls):
        """ Get default user """

        if not cls.__default_user:
            NURESTBasicUser.__default_user = cls()

        return NURESTBasicUser.__default_user

    # Methods

    def prepare_change_password(self, new_password):
        """ Prepares password modification """

        self._new_password = new_password

    def save(self, async=False, callback=None):
        """ Updates the user and perform the callback method """

        if self._new_password:
            self.password = Sha1.encrypt(self._new_password)

        controller = _NURESTSessionCurrentContext.session.login_controller
        controller.password = self._new_password
        controller.api_key = None

        data = json.dumps(self.to_dict())
        request = NURESTRequest(method=HTTP_METHOD_PUT, url=self.get_resource_url(), data=data)

        if async:
            return self.send_request(request=request, async=async, local_callback=self._did_save, remote_callback=callback)
        else:
            connection = self.send_request(request=request)
            return self._did_save(connection)

    def _did_save(self, connection):
        """ Launched when save has been successfully executed """

        self._new_password = None

        controller = _NURESTSessionCurrentContext.session.login_controller
        controller.password = None
        controller.api_key = self.api_key

        if connection.async:
            callback = connection.callbacks['remote']

            if connection.user_info:
                callback(connection.user_info, connection)
            else:
                callback(self, connection)
        else:
            return (self, connection)

    def fetch(self, async=False, callback=None):
        """ Fetch all information about the current object

            Args:
                async (bool): Boolean to make an asynchronous call. Default is False
                callback (function): Callback method that will be triggered in case of asynchronous call

            Returns:
                tuple: (current_fetcher, callee_parent, fetched_bjects, connection)

            Example:
                >>> entity = NUEntity(id="xxx-xxx-xxx-xxx")
                >>> entity.fetch() # will get the entity with id "xxx-xxx-xxx-xxx"
                >>> print entity.name
                "My Entity"
        """
        request = NURESTRequest(method=HTTP_METHOD_GET, url=self.get_resource_url())

        if async:
            return self.send_request(request=request, async=async, local_callback=self._did_fetch, remote_callback=callback)
        else:
            connection = self.send_request(request=request)
            return self._did_retrieve(connection)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015, Alcatel-Lucent Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its contributors
#       may be used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



class NURESTRequest(object):
    """ Request that will be send via the connection """

    def __init__(self, method, url, data=None, params=None, filter=None, page=None, order_by=None):
        """ Initializes a request """

        self._method = method
        self._url = url
        self._data = data
        self._params = params
        self._headers = dict()

        self.set_header('Content-Type', 'application/json')

        if filter:
            self.set_header('X-Nuage-Filter', filter)

        if page:
            self.set_header('X-Nuage-Page', page)

        if order_by:
            self.set_header('X-Nuage-OrderBy', order_by)

    def __str__(self):
        """ Print request """

        return "%s %s" % (self.method, self.url)

    # Properties

    @property
    def method(self):
        """ Get method """

        return self._method

    @method.setter
    def method(self, method):
        """ Set method """

        self._method = method

    @property
    def url(self):
        """ Get url """

        return self._url

    @url.setter
    def url(self, url):
        """ Set url """

        self._url = url

    @property
    def data(self):
        """ Get data """

        return self._data

    @data.setter
    def data(self, data):
        """ Set data """

        self._data = data

    @property
    def params(self):
        """ Get url """

        return self._params

    @params.setter
    def params(self, params):
        """ Set params """

        self._params = params

    @property
    def headers(self):
        """ Prepare headers to send """

        return self._headers

    @headers.setter
    def headers(self, value):
        """ Set header value """

        self._headers = value

    def set_header(self, header, value):
        """ Set header value """

        self._headers[header] = value




# -*- coding: utf-8 -*-
#
# Copyright (c) 2015, Alcatel-Lucent Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its contributors
#       may be used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from time import time


class NURemoteAttribute(object):
    """
        A remote attribute contains information such as:
        :param remote_name: the remote name of the attribute
        :param attribute_type: the type of the attribute
        :param is_required: a boolean to say if the attribute is mandatory or not
        :param is_readonly: a boolean to say if the attribute is read only
        :param max_length: an integer to give the maximum length of the attribute
        :param min_length: an integer to give the minimum length of the attribute
    """

    def __init__(self, local_name, remote_name, attribute_type, min_length=None, max_length=None):
        """ Initializes an attribute """

        self.local_name = local_name
        self.remote_name = remote_name
        self.display_name = local_name
        self.attribute_type = attribute_type
        self.is_required = False
        self.is_readonly = False
        self.is_email = False
        self.is_unique = False
        self.is_editable = True
        self.is_login = False
        self._is_identifier = False
        self.min_length = None
        self.max_length = None
        self._choices = None
        self.has_choices = False
        self._is_password = False
        self.is_forgetable = False

        self.can_order = False
        self.can_search = False

    # Properties

    @property
    def is_identifier(self):
        """ Getter for is_identifier """

        return self._is_identifier

    @is_identifier.setter
    def is_identifier(self, is_identifier):
        """ Setter for is_identifier """

        if is_identifier:
            self.is_editable = False

        self._is_identifier = is_identifier

    @property
    def is_password(self):
        """ Getter for is_identifier """

        return self._is_password

    @is_password.setter
    def is_password(self, is_password):
        """ Setter for is_identifier """

        if is_password:
            self.is_forgetable = True

        self._is_password = is_password

    @property
    def choices(self):
        """ Getter for is_identifier """

        return self._choices

    @choices.setter
    def choices(self, choices):
        """ Setter for is_identifier """

        if choices is not None and len(choices) > 0:
            self.has_choices = True

        self._choices = choices

    # Methods

    def get_default_value(self):
        """ Get a default value of the attribute_type """

        if self.choices:
            return self.choices[0]

        value = self.attribute_type()

        if self.attribute_type is time:
            value = int(value)

        elif self.attribute_type is str:
            value = "A"

        if self.min_length:
            if self.attribute_type is str:
                value = value.ljust(self.min_length, 'a')
            elif self.attribute_type is int:
                value = self.min_length

        elif self.max_length:
            if self.attribute_type is str:
                value = value.ljust(self.max_length, 'a')
            elif self.attribute_type is int:
                value = self.max_length

        return value

    def get_min_value(self):
        """ Get the minimum value """

        value = self.get_default_value()

        if self.attribute_type is str:
            min_value = value[:self.min_length - 1]

        elif self.attribute_type is int:
            min_value = self.min_length - 1

        else:
            raise TypeError('Attribute %s can not have a minimum value' % self.local_name)

        return min_value

    def get_max_value(self):
        """ Get the maximum value """

        value = self.get_default_value()

        if self.attribute_type is str:
            max_value = value.ljust(self.max_length + 1, 'a')

        elif self.attribute_type is int:
            max_value = self.max_length + 1

        else:
            raise TypeError('Attribute %s can not have a maximum value' % self.local_name)

        return max_value



import sys
from thread import get_ident
from .decorators import rewrap, cache_source, classy, decorate

__all__ = [
    'Service', 'replaces', 'setting', 'InputConflict', 'DynamicRuleError',
    'State', 'Action', 'resource', 'registry', 'new', 'empty',
    'lookup', 'manager', 'reraise', 'with_', 'call_with', 'ScopeError',
    'resource_registry',
]

_in_place = """__iadd__ __isub__ __imul__ __idiv__ __itruediv__ __ifloordiv__
__imod__ __ipow__ __ilshift__ __irshift__ __iand__ __ixor__ __ior__""".split()

_ignore = dict.fromkeys("""
    __name__ __module__ __return__ __slots__ get __init__ __metaclass__ __doc__
    __call__ __new__""".split() + _in_place
).__contains__

def _no_in_place(self, *args):
    raise TypeError(
        "In-place operators (other than <<=) cannot be performed on a"
        " service class"
    )

def _ilshift(cls, factory):
    State[cls] = factory
    return cls

def _mod(cls, expr):
    return 'lambda: '+expr

_std_attrs = dict(
    [(k,_no_in_place) for k in _in_place], __ilshift__=_ilshift, __mod__=_mod
)






def redirect_attribute(cls, name, payload):
    setattr(type(cls), name, property(
        lambda s: getattr(s.get(), name),
        lambda s,v: setattr(s.get(), name, v),
        lambda s: delattr(s.get(), name),
    ))


class _ClassDelegate(classy):
    """Type whose attributes/methods are delegated to ``cls.get()``"""

    __slots__ = ()

    get = None  # dummy

    decorate(classmethod)
    def __class_init__(cls, name, bases, cdict, supr):
        meta = type(cls)
        if getattr(meta, '__for_class__', None) is not cls:
            cls.__class__ = meta = type(meta)(
                cls.__name__+'Class', (meta,),
                dict(_std_attrs, __module__=cls.__module__, __for_class__=cls)
            )
            # XXX activate_attrs(meta)?

        supr()(cls, name, bases, cdict, supr)

        if 'get' not in cdict:
            cls.get = staticmethod(classmethod(lookup).__get__(None, cls))

        for k, v in cdict.items():
            if not isinstance(k, basestring):
                continue
            if not isinstance(v, (classmethod,staticmethod))and not _ignore(k):
                redirect_attribute(cls, k, v)






class State(_ClassDelegate):
    """A thread's current configuration and state"""

    def __new__(cls, *rules, **attrs):
        return attrs and object.__new__(cls) or empty()

    def __init__(self, **attrs):
        """Create an empty state with `rules` in effect"""
        self.__dict__.update(attrs)

    def __getitem__(self, key):
        """Get the rule for `key`"""
        return self.getRule(key)

    def __setitem__(self, key, rule):
        """Set the rule for `key`"""
        return self.setRule(key, rule)

    def swap(self):
        """Make this state current and return the old one"""
        raise NotImplementedError("Can't switch to the root state")

    def child(self, *rules):
        """Return a new child state of this one, with `rules` in effect"""
        raise NotImplementedError   # this method is replaced on each instance

    def __enter__(self):
        """Make this state a single-use nested state"""
        raise NotImplementedError("Can't enter the root state")

    def __exit__(self, typ, val, tb):
        """Close this state and invoke exit callbacks"""
        raise NotImplementedError("Can't exit the root state")

    def on_exit(self, callback):
        """Add a `callback(typ,val,tb)` to be invoked at ``__exit__`` time"""
        raise NotImplementedError   # this method is replaced on each instance




    decorate(staticmethod)
    def get(key=None):
        """Return the current state (no args) or a current rule (w/key)"""
        # this method is replaced later below
        raise NotImplementedError

    parent = None


class InputConflict(Exception):
    """Attempt to set a rule that causes a visible conflict in the state"""

class DynamicRuleError(Exception):
    """A fallback or wildcard rule attempted to access dynamic state"""

class ScopeError(Exception):
    """A problem with scoping occurred"""


_exc_info = {}
nones = None, None, None

def _swap_exc_info(data):
    this_thread = get_ident()
    old = _exc_info.get(this_thread, nones)
    _exc_info[this_thread] = data
    return old


def new():
    """Return a new child of the current state"""
    return State.child()









def _let_there_be_state():
    """Create a world of states, manipulable only by exported functions"""

    states = {}

    def _swap(what):
        this_thread = get_ident()
        old = states.setdefault(this_thread, what)
        states[this_thread] = what
        return old

    def lookup(key):
        """Return the value of `key` in the current state"""
        try:
            state, getRule, lookup, child = states[get_ident()]
        except KeyError:
            empty().swap()
            state, getRule, lookup, child = states[get_ident()]
        return lookup(key)

    def get(key=None):
        try:
            state, getRule, lookup, child = states[get_ident()]
        except KeyError:
            empty().swap()
            state, getRule, lookup, child = states[get_ident()]
        if key is None:
            return state
        return getRule(key)

    def disallow(key):
        raise DynamicRuleError(
            "default rule or exit function tried to read dynamic state", key
        )

    def empty():
        """Return a new, empty State instance"""
        state = new_state(root_getrule)
        state.parent = root
        return state

    def new_state(inherit=None, inheritedDistances=None, propagate=None):

        buffer = {}
        rules = {}
        values = {}
        distances = {}
        computing = {}

        get_stack, set_stack = computing.get, computing.setdefault

        def getRule(key):
            """Get my rule for `key`"""
            try:
                rule = rules[key]
            except KeyError:
                try:
                    rule = buffer[key]
                except KeyError:
                    rule = buffer.setdefault(key, __fallback(key))

                # snapshot the currently-set value - this is thread-safe
                # because setdefault is atomic, so rules[key] will only *ever*
                # have one value, even if it's not the one now in the buffer
                #
                rule = rules.setdefault(key, rule)
                if key not in distances:
                    if inheritedDistances is not None and inherit(key)==rule:
                        distances.setdefault(key, inheritedDistances[key]+1)
                    else:
                        distances.setdefault(key, 0)

            if computing:
                # Ensure that any value being computed in this thread has a
                # maximum propagation distance no greater than this rule's
                # distance.
                stack = get_stack(get_ident())
                if stack:
                    stack[-1] = min(stack[-1], distances[key])

            return rule

        def setRule(key, rule):
            """Set my rule for `key`, or raise an error if inconsistent"""
            buffer[key] = rule
            # as long as a snapshot hasn't been taken yet, `old` will be `rule`
            old = rules.get(key, rule)
            if old is not rule and old != rule:
                raise InputConflict(key, old, rule)

        def getValue(key):
            """Get the dynamic value of `key`"""
            try:
                value = values[key]

            except KeyError:
                this_thread = get_ident()
                rule = getRule(key)     # this ensures distances[key] is known

                stack = set_stack(this_thread, [])
                stack.append(distances[key])

                try:
                    value = key.__apply__(key, rule)
                finally:
                    distance = stack.pop()
                    if not stack:
                        del computing[this_thread]
                    else:
                        stack[-1] = min(stack[-1], distance)

                value = publish(distance, key, value)

            else:
                if computing:
                    stack = get_stack(get_ident())
                    if stack:
                        stack[-1] = min(stack[-1], distances[key])

            return value



        def publish(distance, key, value):
            """Accept value from this state or child, and maybe propagate it"""

            # It's safe to update distances here because no thread can depend
            # on the changed distance for propagation unless *some* thread has
            # already finished the relevant computation -- i.e., done this very
            # update.  Otherwise, there would be no values[key], therefore the
            # thread would have to follow the same code path, ending up by
            # being the thread doing this update!  Ergo, this is a safe update.
            #
            distances[key] = distance

            if distance and propagate:
                # pass it up to the parent, but use the value the parent has.
                # therefore, the parent at "distance" height from us will be
                # the arbiter of the value for all its children, ensuring that
                # exactly one value is used for the entire subtree!
                #
                value = propagate(distance-1, key, value)

            # Return whatever value wins at this level to our children
            return values.setdefault(key, value)


        def child():
            """Return a new child state"""
            s = new_state(getRule, distances, publish)
            s.parent = this
            return s


        def __fallback(key):
            """Compute the fallback for key"""
            old = _swap(disabled)
            try:
                return key.__fallback__(inherit, key)
            finally:
                _swap(old)



        def swap():
            if exited:
                raise ScopeError("Can't switch to an exited state")
            state, get, lookup, old_child = old = _swap(enabled)
            if lookup is disallow:
                _swap(old)
                raise DynamicRuleError(
                    "default rule or exit function tried to change states"   # XXX
                )
            return state

        def __enter__():
            if my_parent or exited:
                raise ScopeError("Can't re-enter a previously-entered state")
            elif active_child:
                raise ScopeError("State already has an active child")
            elif get() is this:
                raise ScopeError("State is already current")
            parent, xx, xx, parents_child = old = states[get_ident()]
            if parents_child:
                raise ScopeError("Current state already has an active child")
            swap()
            my_parent.append(old); parents_child.append(this)
            return this

        def __exit__(typ, val, tb):
            if exited:
                raise ScopeError("State already exited")
            elif not my_parent:
                raise ScopeError("State hasn't been entered yet")
            elif active_child:
                raise ScopeError("Nested state(s) haven't exited yet")
            elif get() is not this:
                raise ScopeError("Can't exit a non-current state")
            parents_child = my_parent[0][-1]
            _swap(my_parent.pop()) # reactivate parent state
            parents_child.pop()
            exited.append(1)
            values.clear()
            return call_exitfuncs(typ, val, tb)

        active_child = []
        my_parent = []
        exited = []
        exit_functions = []

        def call_exitfuncs(typ, val, tb):
            old = _swap(disabled)
            try:
                for func in exit_functions:
                    try:
                        func(typ, val, tb)
                    except:
                        typ, val, tb = sys.exc_info()
            finally:
                _swap(old)
                del typ, val, tb

        def on_exit(callback):
            if exited:
                raise ScopeError("State already exited")
            elif not my_parent:
                raise ScopeError("State hasn't been entered yet")
            if callback not in exit_functions:
                exit_functions.append(callback)

        this = State(
            getRule=getRule, setRule=setRule, swap=swap, child=child,
            __enter__=__enter__, __exit__=__exit__, on_exit = on_exit
        )
        enabled  = this, getRule, getValue, active_child
        disabled = None, getRule, disallow, active_child
        return this

    State.get = staticmethod(get)
    State.root = root = new_state(); root.child = empty
    root_getrule = root.getRule
    del root.swap, root.__enter__, root.__exit__
    return lookup, empty

lookup, empty = _let_there_be_state(); del _let_there_be_state

class _GeneratorContextManager(object):
    """Helper for @context.manager decorator."""

    def __init__(self, gen):
        self.gen = gen

    def __enter__(self):
        for value in self.gen:
            return value
        else:
            raise RuntimeError("generator didn't yield")

    def __exit__(self, typ, value, traceback):
        if typ is None:
            for value in self.gen:
                raise RuntimeError("generator didn't stop")
        else:
            try:
                old = _swap_exc_info((typ, value,traceback))
                try:
                    self.gen.next()
                finally:
                    _swap_exc_info(old)

                raise RuntimeError("generator didn't stop after throw()")

            except StopIteration, exc:
                # Suppress the exception *unless* it's the same exception that
                # was passed to throw().  This prevents a StopIteration
                # raised inside the "with" statement from being suppressed
                return exc is not value
            except:
                # only re-raise if it's *not* the exception that was
                # passed to throw(), because __exit__() must not raise
                # an exception unless __exit__() itself failed.  But throw()
                # has to raise the exception to signal propagation, so this
                # fixes the impedance mismatch between the throw() protocol
                # and the __exit__() protocol.
                if sys.exc_info()[1] is not value:
                    raise

def manager(func):
    """Emulate 2.5 ``@contextlib.contextmanager`` decorator"""
    gcm = _GeneratorContextManager
    return rewrap(func, lambda *args, **kwds: gcm(func(*args, **kwds)))

def with_(ctx, func):
    """Perform PEP 343 "with" logic for Python versions <2.5

    The following examples do the same thing at runtime::

        Python 2.5+          Python 2.3/2.4
        ------------         --------------
        with x as y:         z = with_(x,f)
            z = f(y)

    This function is used to implement the ``call_with()`` decorator, but
    can also be used directly.  It's faster and more compact in the case where
    the function ``f`` already exists.
    """
    inp = ctx.__enter__()
    try:
        retval = func(inp)
    except:
        if not ctx.__exit__(*sys.exc_info()):
            raise
    else:
        ctx.__exit__(None, None, None)
        return retval

def reraise():
    """Reraise the current contextmanager exception, if any"""
    typ,val,tb = _exc_info.get(get_ident(), nones)
    if typ:
        try:
            raise typ,val,tb
        finally:
            del typ,val,tb




def call_with(ctxmgr):
    """Emulate the PEP 343 "with" statement for Python versions <2.5

    The following examples do the same thing at runtime::

        Python 2.5+          Python 2.4
        ------------         -------------
        with x as y:         @call_with(x)
            print y          def do_it(y):
                                 print y

    ``call_with(foo)`` returns a decorator that immediately invokes the
    function it decorates, passing in the same value that would be bound by
    the ``as`` clause of the ``with`` statement.  Thus, by decorating a
    nested function, you can get most of the benefits of "with", at a cost of
    being slightly slower and perhaps a bit more obscure than the 2.5 syntax.

    Note: because of the way decorators work, the return value (if any) of the
    ``do_it()`` function above will be bound to the name ``do_it``.  So, this
    example prints "42"::

        @call_with(x)
        def do_it(y):
            return 42

        print do_it

    This is rather ugly, so you may prefer to do it this way instead, which
    more explicitly calls the function and gets back a value::

        def do_it(y):
            return 42

        print with_(x, do_it)
    """
    return with_.__get__(ctxmgr, type(ctxmgr))





noop = lambda key, rule: rule
mngr = lambda key, rule: Action.manage(rule())
call = lambda key, rule: rule()

class Service(_ClassDelegate):
    """A replaceable, thread-local singleton"""

    __slots__ = ()  # pure mixin class

    def __default__(cls):
        return cls()

    __default__ = classmethod(__default__)

    def __fallback__(cls, inherit, key):
        if inherit: return inherit(key)
        return cls.__default__

    __fallback__ = classmethod(__fallback__)
    __apply__ = staticmethod(call)

    def new(cls, factory=None):
        factory = factory or cls
        state = State.child().__enter__()
        try:
            # we use a lambda below to ensure that the rule is unique; that way,
            # we are guaranteed to get a *new* instance of the service in the
            # scope, rather than reusing the existing one.
            state[cls.get.im_self] = lambda: factory()
            yield cls.get()
            reraise()
        except:
            state.__exit__(*sys.exc_info())
            raise
        else:
            state.__exit__(None, None, None)

    new = classmethod(manager(new))



class Scope(Service):
    """A scope for resources"""

    def __init__(self):
        self.state = State.get()
        self.cache = {}

    def __compute__(cls, key, func, input):
        self = cls.get()
        cache = self.cache
        if cache is None:
            raise ScopeError(self.__class__.__name__+" is already exited")
        elif input is not self.state[key]:
            raise ScopeError("Redefined rule in sub-state")
        elif key in cache:
            return cache[key]
        else:
            old = self.state.swap()
            try:
                cache[key] = value = self.manage(func(input))
                self.state.on_exit(self.atexit)
                return value
            finally:
                old.swap()

    __compute__ = classmethod(__compute__)

    def manage(self, ob):
        """Do any necessary magic on 'ob' and return the new value"""
        return ob

    def atexit(self, *exc):
        self.cache = None

    def __default__(cls):
        raise RuntimeError("No %s is currently active" % (cls.__name__))
    __default__ = classmethod(__default__)




    def resource(cls, func):
        """Decorator to create a scoped resource"""
        return setting(func, wrap=cls.__compute__)
    resource = classmethod(resource)

    def resource_registry(cls, func):
        """Decorator to create a scoped resource registry"""
        return registry(func, wrap=cls.__compute__)
    resource_registry = classmethod(resource_registry)


class Action(Scope):
    """Service for managing transaction-scoped resources"""

    def __init__(self):
        self.managers = []
        super(Action, self).__init__()

    def atexit(self, *exc):
        super(Action, self).atexit(*exc)
        managers, self.managers = self.managers, None
        while managers:
            managers.pop().__exit__(*exc)  # XXX how do we handle errors?

    def manage(self, ob):
        enter = getattr(ob, '__enter__', None)
        if enter is None:
            return ob
        ctx = ob
        ob  = ctx.__enter__()

        # don't call __exit__ unless __enter__ succeeded
        # (if there was an error, we wouldn't have gotten this far)
        self.managers.append(ctx)
        return ob
        # XXX return self.manage(ob)

resource = Action.resource
resource_registry = Action.resource_registry


def nop(): pass

class setting(object):
    """Decorator that turns a function into a contextual variable"""

    __class__ = type(nop)
    func_code = nop.func_code
    func_defaults = ()

    __argnames__ = ('value','expr'),

    def __init__(self, func, wrap=None):
        if func.func_code.co_argcount != len(self.__argnames__):
            raise TypeError(
                type(self).__name__+" function must have exactly %d argument(s)"
                % len(self.__argnames__)
            )
        for num, (var, names) in enumerate(
            zip(func.func_code.co_varnames, self.__argnames__)
        ):
            if var not in names:
                raise TypeError(
                    type(self).__name__+" function argument %d must be named '%s'"
                    % (num+1, "' or '".join(names))
                )
        if self.__argnames__ and (not func.func_defaults or len(func.func_defaults)!=1):
            raise TypeError(
                type(self).__name__ +
                " function must have a default value for last argument"
            )

        self.__function__  = func
        self.__module__    = func.__module__
        self.__name__      = func.__name__
        self.__doc__       = func.__doc__
        if wrap: self.__wrap__ = wrap

    __call__ = lookup



    def __apply__(self, key, input):
        return self.__wrap__(key, self.__function__, input)

    def __wrap__(self, key, func, input):
        return func(input)

    def __fallback__(self, inherit, key):
        if inherit is None:
            return self.__function__.func_defaults[0]
        else:
            return inherit(key)

    def __ilshift__(self, value):
        State[self] = value
        return self

    def __repr__(self):
        if self.__module__: return self.__module__ + '.' + self.__name__
        return self.__name__

    def __mod__(self, other):
        if self.__function__.func_code.co_varnames[
            len(type(self).__argnames__)-1
        ]=='expr':
            return 'lambda: '+other
        return other















def _with_prefix((pre, func), suffix):
    return func(pre+suffix)

def _prefixer(prefix, func):
    return _with_prefix.__get__((prefix, func), tuple)

if _prefixer(1, 2) != _prefixer(1, 2):  # 2.3 and 2.4 don't compare selves
    class _prefixer(tuple):
        __slots__ = []
        def __new__(cls, prefix, func):
            return tuple.__new__(cls, (prefix, func))
        def __init__(cls, prefix, func):
            pass
        def __call__(self, suffix):
            return self[1](self[0]+suffix)

class wildcard(setting):
    """Object used to do parent namespace rule lookups"""

    def __init__(self, registry):
        self.__module__ = registry.__module__
        self.__name__   = registry.__name__ + '.*'
        self.__function__ = registry.__function__
        self.__namespace__ = registry

    def __mod__(self, other):
        return 'lambda suffix: '+other

    def __apply__(self, key, input):
        return input

    def __fallback__(self, inherit, key):
        parent = self.__namespace__
        if parent.__namespace__:
            func = State.get(parent.__namespace__['*'])
            if func is not None:
                prefix = self.__name__.split('.')[-2]+'.'
                return _prefixer(prefix, func)
        # couldn't find a parent wildcard rule, return None


class registry(setting):
    """Decorator that turns a function into a contextual registry"""

    __argnames__ = ('suffix',), ('value','expr')
    func_code = (lambda key, default: None).func_code
    func_defaults = (None,)

    def __init__(self, func, ns=None, name='', wrap=None):
        setting.__init__(self, func, wrap)
        self.__name__ = name or self.__name__
        self.__namespace__ = ns
        self.__contents__ = {'*': wildcard(self)}

    def __getitem__(self, key):
        if '.' in key:
            for key in key.split('.'): self = self[key]
            return self
        try:
            return self.__contents__[key]
        except KeyError:
            s = self.__dict__[key] = self.__contents__.setdefault(key, registry(
                self.__function__, self, self.__name__ + '.' + key,
                self.__dict__.get('__wrap__')
            ))
            return s

    def __getattr__(self, key):
        if key.startswith('__') and key.endswith('__'):
            raise AttributeError(key)
        return self[key]

    def __contains__(self,key):
        if '.' in key:
            for key in key.split('.'):
                if key not in self.__contents__:
                    return False
                self = self[key]
            else:
                return True
        return key in self.__contents__

    def __iter__(self):
        return iter([key for key in self.__contents__ if key!='*'])

    def __call__(self, key, default=None):
        if key not in self:
            return default
        return lookup(self[key])

    def __apply__(self, key, input):
        return self.__wrap__(key, self.__function__.__get__(''), input)

    def __fallback__(self, inherit, key):
        if self.__namespace__:
            # Not the root registry, try looking up wildcard rule(s)
            suffix = key.__name__[len(self.__namespace__.__name__)+1:]
            finder = State.get(self.__namespace__['*'])
            if finder is not None:
                return finder(suffix)

        # No wildcards in registries above me, so try to inherit instead:
        if inherit is not None:
            return inherit(key)

        # We're in the root state; make sure we're in the root registry too
        while self.__namespace__: self = self.__namespace__

        # And then invoke the registry function with the suffix
        suffix = key.__name__[len(self.__name__)+1:]
        return self.__function__(suffix)

    def __setitem__(self, key, value):
        if value is not self[key]:
            raise TypeError("Registries are read-only")

    def __setattr__(self, key, value):
        if key.startswith('__') and key.endswith('__'):
            return object.__setattr__(self, key, value)
        else:
            self[key] = value


class Source(object):
    """Object representing a source file (or pseudo-file)"""

    __slots__ = "filename", "__weakref__"

    def __init__(self, filename, source=None):
        global linecache; import linecache
        self.filename = filename
        if source is not None:
            cache_source(filename, source, self)

    def compile(self, *args, **kw):
        return Line(''.join(self), self, 1).compile(*args, **kw)

    def __getitem__(self, key):
        return Line(linecache.getlines(self.filename)[key], self, key+1)

    def __repr__(self):
        return "Source(%r)" % self.filename

    def recode(self, code, offset=0):
        import new
        if not isinstance(code, new.code):
            return code
        return new.code(
            code.co_argcount, code.co_nlocals, code.co_stacksize,
            code.co_flags, code.co_code,
            tuple([self.recode(c, offset) for c in code.co_consts]+[self]),
            code.co_names, code.co_varnames, code.co_filename, code.co_name,
            code.co_firstlineno+offset, code.co_lnotab, code.co_freevars,
            code.co_cellvars
        )









class Line(str):
    """String that knows its file and line number"""

    def __new__(cls, text, source, line):
        return str.__new__(cls, text)

    def __init__(self, text, source, line):
        self.source = source
        self.line = line

    def compile(self, *args, **kw):
        # XXX needs syntax error trapping, unicode encoding support
        code = compile(self, self.source.filename, *args, **kw)
        return self.source.recode(code, self.line-1)

    def eval(self, *args):
        return eval(self.compile('eval'), *args)

    def splitlines(self, *args, **kw):
        return [Line(line, self.source, self.line+offset)
            for offset, line in enumerate(str.splitlines(self, *args, **kw))]

    for m in [
        'capitalize', 'center', 'expandtabs', 'ljust', 'lower', 'lstrip',
        'replace', 'rjust', 'rstrip', 'strip', 'swapcase', 'title',
        'translate', 'upper', 'zfill', '__add__', '__radd__', '__getslice__',
        '__mod__', '__rmod__',
    ]:
        if hasattr(str, m):
            locals()[m] = (lambda f:
                lambda self,*args,**kw: Line(
                    f(self,*args,**kw), self.source, self.line
                )
            )(getattr(str,m))







def replaces(target):
    """Class decorator to indicate that this service replaces another"""

    def decorator(cls):
        if not issubclass(cls, Service):
            raise TypeError(
                "context.replaces() can only be used in a context.Service"
                " subclass"
            )
        cls.get = staticmethod(target.get)
        return cls

    from peak.util.decorators import decorate_class
    decorate_class(decorator)

    # Ensure that context.replaces() is used only once per class suite
    cdict = sys._getframe(1).f_locals

    if cdict.setdefault('get', target.get) is not target.get:
        raise ValueError(
            "replaces() must be used only once per class;"
            " there is already a value for ``get``: %r"
            % (cdict['get'],)
        )




















# -*- coding: utf-8 -*-
#
# Copyright (c) 2015, Alcatel-Lucent Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its contributors
#       may be used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from base64 import urlsafe_b64encode


class NURESTLoginController(object):
    """ Holds information about the current user session """

    def __init__(self):
        """ Initiliazes the login controller """

        if not hasattr(self, '_initiliazed') or not self._initiliazed:
            self._initiliazed = True
            self._is_impersonating = False
            self._impersonation = None

            self._certificate = None
            self._user = None
            self._password = None
            self._api_key = None
            self._enterprise = None
            self._url = None
            self._async = True

    def __str__(self):
        """ Prints information """

        return "[NURESTLoginController] user=%s enterprise=%s url=%s (API Key=%s)" % (self._user, self._enterprise, self._url, self._api_key)

    # Properties

    @property
    def is_impersonating(self):
        """ Check if we are currently using impersonation

            Returns:
                Returns True if currently using impersonation
        """

        return self._is_impersonating


    @property
    def impersonation(self):
        """ Get impersonation information

            Returns:
                Returns string information about the impersonsation
        """
        return self._impersonation

    @property
    def user(self):
        """ Get the user name

            Returns:
                Returns the name of the user
        """

        return self._user

    @user.setter
    def user(self, user):
        """ Set the user name to connect with

            Args:
                user: name of the user
        """

        self._user = user

    @property
    def password(self):
        """ Get user password

            Returns:
                Returns user password
        """

        return self._password

    @password.setter
    def password(self, password):
        """ Set user password to connect with

            Args:
                password: the password of the user
        """

        self._password = password

    @property
    def certificate(self):
        """ Get the certificate used for authentication

            Returns:
                Returns the certificate
        """
        return self._certificate

    @certificate.setter
    def certificate(self, certificate):
        """ Set certificate to connect with

            Args:
                certificate: the certificate
        """
        self._certificate = certificate

    @property
    def api_key(self):
        """ Get API Key

            API Key is available when NURESTUser has been fetched.

            Returns:
                Returns the string containing the API Key
        """

        return self._api_key

    @api_key.setter
    def api_key(self, api_key):
        """ Set the API Key

            When set, the API key will be used to make all HTTP calls

            Args:
                api_key: the string that represents the API Key
        """

        self._api_key = api_key

    @property
    def enterprise(self):
        """ Get the enterprise name

            Returns:
                Returns the name of the enterprise
        """

        return self._enterprise

    @enterprise.setter
    def enterprise(self, enterprise):
        """ Set the enterprise name

            Args:
                enterprise: the name of the enterprise
        """

        self._enterprise = enterprise

    @property
    def url(self):
        """ Get API URL endpoint

            Returns:
                Returns the API URL endpoint
        """
        return self._url

    @url.setter
    def url(self, url):
        """ Set API URL endpoint

            Args:
                url: the url of the API endpoint
        """
        if url and url.endswith('/'):
            url = url[:-1]

        self._url = url

    @property
    def async(self):
        """ Is asynchronous controller

            Returns:
                Returns True if the controller is asynchronous
        """

        return self._async

    @async.setter
    def async(self, async):
        """ Set asynchronous controller

            Args:
                async: Boolean to say whether or not the controller is async.
        """
        self._async = async


    # Methods

    def get_authentication_header(self, user=None, api_key=None, password=None, certificate=None):
        """ Return authenication string to place in Authorization Header

            If API Token is set, it'll be used. Otherwise, the clear
            text password will be sent. Users of NURESTLoginController are responsible to
            clean the password property.

            Returns:
                Returns the XREST Authentication string with API Key or user password encoded.
        """

        if not user:
            user = self.user

        if not api_key:
            api_key = self.api_key

        if not password:
            password = self.password

        if not password:
            password = self.password

        if not certificate:
            certificate = self._certificate

        if certificate:
            return "XREST %s" % urlsafe_b64encode("%s:%s" % (user, ""))

        if api_key:
            return "XREST %s" % urlsafe_b64encode("%s:%s" % (user, api_key))

        if isinstance(password, unicode):
            password = password.encode("utf-8")

        return "XREST %s" % urlsafe_b64encode("%s:%s" % (user, password))

    def reset(self):
        """ Reset controller

            It removes all information about previous session
        """

        self._is_impersonating = False
        self._impersonation = None

        self.user = None
        self.password = None
        self.api_key = None
        self.enterprise = None
        self.url = None

    def impersonate(self, user, enterprise):
        """ Impersonate a user in a enterprise

            Args:
                user: the name of the user to impersonate
                enterprise: the name of the enterprise where to use impersonation
        """

        if not user or not enterprise:
            raise ValueError('You must set a user name and an enterprise name to begin impersonification')

        self._is_impersonating = True
        self._impersonation = "%s@%s" % (user, enterprise)

    def stop_impersonate(self):
        """ Stop impersonization """

        if self._is_impersonating:
            self._is_impersonating = False
            self._impersonation = None

    def equals(self, controller):
        """ Verify if the controller corresponds
            to the current one.

        """
        if controller is None:
            return False

        return self.user == controller.user and self.enterprise == controller.enterprise and self.url == controller.url



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015, Alcatel-Lucent Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its contributors
#       may be used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import json
import logging

from copy import deepcopy

from time import time

from .exceptions import BambouHTTPError, InternalConsitencyError
from .nurest_connection import NURESTConnection, HTTP_METHOD_DELETE, HTTP_METHOD_PUT, HTTP_METHOD_POST, HTTP_METHOD_GET
from .nurest_request import NURESTRequest
from .nurest_session import _NURESTSessionCurrentContext
from .utils import NURemoteAttribute

from bambou import bambou_logger
from bambou.config import BambouConfig


class NUMetaRESTObject(type):

    @property
    def rest_name(cls):
        if cls.__name__ == "NURESTBasicUser" or cls.__name__ == "NURESTObject":
            return "Not Implemented"

        if cls.__rest_name__ is None:
            raise NotImplementedError('%s has no defined name. Implements rest_name property first.' % cls)
        return cls.__rest_name__

    @property
    def rest_resource_name(cls):
        rest_name = cls.rest_name

        if cls.is_resource_name_fixed():
            return rest_name

        last_letter = rest_name[-1]

        if last_letter == "y":

            vowels = ['a', 'e', 'i', 'o', 'u', 'y']

            if rest_name[-2].lower() not in vowels:
                rest_name = rest_name[:len(rest_name) - 1]
                rest_name += "ies"
            else:
                rest_name += "s"

        elif last_letter != "s":
            rest_name += "s"

        return rest_name


class NURESTObject(object):
    """ Determines an object as a NURESTObject one
        Provides basic saving and fetching utilities
    """

    __metaclass__ = NUMetaRESTObject
    __rest_name__ = None

    def __init__(self):
        """ Initializes the object with general information

            Args:
                creation_date: datetime when the object as been created
                external_id: external identifier of the object
                id: identifier of the object
                local_id: internal identifier of the object
                owner: string representing the owner
                parent_id: identifier of the object's parent
                parent_type: type of the parent
        """

        self._creation_date = None
        self._external_id = None
        self._id = None
        self._local_id = None
        self._owner = None
        self._parent_id = None
        self._parent_type = None
        self._parent = None
        self._last_updated_by = None
        self._last_updated_date = None
        self._is_dirty = False

        self._attribute_errors = dict()
        self._attributes = dict()

        self.expose_attribute(local_name=u'id', remote_name=u'ID', attribute_type=str, is_identifier=True)
        self.expose_attribute(local_name=u'parent_id', remote_name=u'parentID', attribute_type=str)
        self.expose_attribute(local_name=u'parent_type', remote_name=u'parentType', attribute_type=str)
        self.expose_attribute(local_name=u'creation_date', remote_name=u'creationDate', attribute_type=time, is_editable=False)
        self.expose_attribute(local_name=u'owner', attribute_type=str, is_readonly=True)
        self.expose_attribute(local_name=u'last_updated_date', remote_name=u'lastUpdatedDate', attribute_type=time, is_readonly=True)
        self.expose_attribute(local_name=u'last_updated_by', remote_name=u'lastUpdatedBy', attribute_type=str, is_readonly=True)

        self._fetchers_registry = dict()

    def _compute_args(self, data=dict(), **kwargs):
        """ Compute the arguments

            Try to import attributes from data.
            Otherwise compute kwargs arguments.

            Args:
                data: a dict()
                kwargs: a list of arguments
        """

        for name, remote_attribute in self._attributes.items():
            default_value = BambouConfig.get_default_attribute_value(self.__class__, name, remote_attribute.attribute_type)
            setattr(self, name, default_value)

        if len(data) > 0:
            self.from_dict(data)

        for key, value in kwargs.iteritems():
            if hasattr(self, key):
                setattr(self, key, value)

    # Properties

    @property
    def creation_date(self):
        """ Get creation date """

        return self._creation_date

    @creation_date.setter
    def creation_date(self, creation_date):
        """ Set creation date """

        self._creation_date = creation_date

    @property
    def external_id(self):
        """ Get external id """

        return self._external_id

    @external_id.setter
    def external_id(self, external_id):
        """ Set external id """

        self._external_id = external_id

    @property
    def id(self):
        """ Get object id """

        return self._id

    @id.setter
    def id(self, id):
        """ Set object id """

        self._id = id

    @property
    def local_id(self):
        """ Get local id """

        return self._local_id

    @local_id.setter
    def local_id(self, local_id):
        """ Set local id """

        self._local_id = local_id

    @property
    def owner(self):
        """ Get owner """

        return self._owner

    @owner.setter
    def owner(self, owner):
        """ Set owner """

        self._owner = owner

    @property
    def parent_id(self):
        """ Get parent id """

        return self._parent_id

    @parent_id.setter
    def parent_id(self, parent_id):
        """ Set parent id """

        self._parent_id = parent_id

    @property
    def parent_type(self):
        """ Get parent type """

        return self._parent_type

    @parent_type.setter
    def parent_type(self, parent_type):
        """ Set parent type """

        self._parent_type = parent_type

    @property
    def parent_object(self):
        """ Get parent """

        return self._parent

    @parent_object.setter
    def parent_object(self, parent):
        """ Set parent id """

        self._parent = parent

    @property
    def last_updated_by(self):
        """ Get last updated by user id info """

        return self._last_updated_by

    @last_updated_by.setter
    def last_updated_by(self, user_id):
        """ Set last updated by user id info """

        self._last_updated_by = user_id

    @property
    def last_updated_date(self):
        """ Get last updated date """

        return self._last_updated_date

    @last_updated_date.setter
    def last_updated_date(self, update_date):
        """ Set last updated by user id info """

        self._last_updated_date = update_date

    @property
    def rest_name(self):
        """ Returns the current ReST name of the object.

            Returns:
                Returns a dictionnary containing attribute information
        """
        return self.__class__.rest_name

    @property
    def rest_resource_name(self):
        """ Resource name of the object.

            It will compute the plural if needed

            Returns:
                Returns a string that represents the resouce name of the object
        """
        return self.__class__.rest_resource_name

    @property
    def fetchers(self):
        """ Return a copy of all fetchers

            Returns:
                list: list of all fetchers

            Example:
                >>> print entity.fetchers
                [<NUSubEntitiesFetcher at xxxx>, <NUOtherEntitiesFetcher at yyyy>]
        """
        return deepcopy(self._fetchers_registry.values())

    # Children

    @property
    def children_rest_names(self):
        """ Gets the list of all possible children ReST names.

            Returns:
                list: list containing all possible rest names as string

            Example:
                >>> entity = NUEntity()
                >>> entity.children_rest_names
                ["foo", "bar"]
        """

        names = []

        for fetcher in self.fetchers:
            names.append(fetcher.__class__.managed_object_rest_name())

        return names

    # Class methods

    @classmethod
    def rest_base_url(cls):
        """ Override this method to set object base url """

        controller = _NURESTSessionCurrentContext.session.login_controller
        return controller.url

    @classmethod
    def is_resource_name_fixed(cls):
        """ Boolean to say if the resource name should be fixed. Default is False """

        return False

    @classmethod
    def object_with_id(cls, id):
        """ Get a new object with the given id

            Returns:
                Returns a new instance with specified id
        """

        new_object = cls()
        new_object.id = id

        return new_object

    # URL and resource management

    def get_resource_url(self):
        """ Get resource complete url """

        name = self.__class__.rest_resource_name
        url = self.__class__.rest_base_url()

        if self.id is not None:
            return "%s/%s/%s" % (url, name, self.id)

        return "%s/%s" % (url, name)

    def get_resource_url_for_child_type(self, nurest_object_type):
        """ Get the resource url for the nurest_object type """

        return "%s/%s" % (self.get_resource_url(), nurest_object_type.rest_resource_name)

    def __str__(self):
        """ Prints a NURESTObject """

        return "%s (ID=%s)" % (self.__class__, self.id)

    def validate(self):
        """ Validate the current object attributes.

            Check all attributes and store errors

            Returns:
                Returns True if all attibutes of the object
                respect contraints. Returns False otherwise and
                store error in errors dict.

        """
        self._attribute_errors = dict()  # Reset validation errors

        for local_name, attribute in self._attributes.iteritems():

            value = getattr(self, local_name, None)

            if value is None and attribute.is_required:
                self._attribute_errors[local_name] = 'Attribute %s is required' % (local_name)
                continue

            if value is None:
                continue  # without error

            if type(value) != attribute.attribute_type:
                if attribute.attribute_type != str or type(value) != unicode:
                    self._attribute_errors[local_name] = 'Attribute %s is not of type %s' % (local_name, attribute.attribute_type)
                    continue

            if attribute.min_length and len(value) < attribute.min_length:
                self._attribute_errors[local_name] = 'Attribute %s length is too short (minimum=%s)' % (local_name, attribute.min_length)
                continue

            if attribute.max_length and len(value) > attribute.max_length:
                self._attribute_errors[local_name] = 'Attribute %s length is too long (maximum=%s)' % (local_name, attribute.max_length)
                continue

            if attribute.choices and value not in attribute.choices:
                self._attribute_errors[local_name] = 'Attribute %s is not a valid option (choices=%s)' % (local_name, attribute.choices)
                continue

        return self.is_valid()

    def is_valid(self):
        """
        """
        return len(self.errors) == 0

    @property
    def errors(self):
        """
        """
        return self._attribute_errors


    def expose_attribute(self, local_name, attribute_type, remote_name=None, display_name=None, is_required=False, is_readonly=False, max_length=None, min_length=None, is_identifier=False, choices=None, is_unique=False, is_email=False, is_login=False, is_editable=True, is_password=False, can_order=False, can_search=False):
        """ Expose local_name as remote_name

            An exposed attribute `local_name` will be sent within the HTTP request as
            a `remote_name`

        """
        if remote_name is None:
            remote_name = local_name

        if display_name is None:
            display_name = local_name

        attribute = NURemoteAttribute(local_name=local_name, remote_name=remote_name, attribute_type=attribute_type)
        attribute.display_name = display_name
        attribute.is_required = is_required
        attribute.is_readonly = is_readonly
        attribute.min_length = min_length
        attribute.max_length = max_length
        attribute.is_editable = is_editable
        attribute.is_identifier = is_identifier
        attribute.choices = choices
        attribute.is_unique = is_unique
        attribute.is_email = is_email
        attribute.is_login = is_login
        attribute.is_password = is_password
        attribute.can_order = can_order
        attribute.can_search = can_search

        self._attributes[local_name] = attribute

    def get_attributes(self):
        """ Get all attributes information

            Returns:
                Returns a dictionnary containing attribute information
        """

        return self._attributes.values()

    def get_attribute_infos(self, local_name):
        """ Get exposed attribute information

            Args:
                local_name: the attribute name

            Returns:
                A dictionary of all information

        """
        if local_name in self._attributes:
            return self._attributes[local_name]

        return None

    def is_owned_by_current_user(self):
        """ Check if the current user owns the object """

        from bambou.nurest_user import NURESTBasicUser
        current_user = NURESTBasicUser.get_default_user()
        return self._owner == current_user.id

    def parent_for_matching_rest_name(self, rest_names):
        """ Return parent that matches a rest name """

        parent = self

        while parent:
            if parent.rest_name in rest_names:
                return parent

            parent = parent.parent

        return None

    def _genealogic_types(self):
        """ Get genealogic types

            Returns:
                Returns a list of all parent types
        """

        types = []
        parent = self

        while parent:
            types.push(parent.rest_name)
            parent = parent.parent

        return types

    def _genealogic_ids(self):
        """ Get all genealogic ids

            Returns:
                 A list of all parent ids
        """

        ids = []
        parent = self

        while parent:
            ids.push(parent.id)
            parent = parent.parent

        return ids

    def _genealogy_contains_type(self, resource_name):
        """ Check if parents contains an object of type resource_name

            Args:
                resource_name: the name of the resource to find

            Returns:
                Returns True if a parent of type has been found. False otherwise.
        """

        resource_names = self._genealogic_types()
        return resource_name in resource_names

    def _genealogy_contains_id(self, id):
        """ Check if parents contains an object of type resource_name

            Args:
                id: the id of the resource to find

            Returns:
                Returns True if a parent with specific id has been found. False otherwise.
        """

        ids = self._genealogic_ids()
        return id in ids

    def get_formated_creation_date(self, format='mmm dd yyyy HH:MM:ss'):
        """ Return creation date with a given format. Default is 'mmm dd yyyy HH:MM:ss' """

        if not self._creation_date:
            return u"No date"

        return self._creation_date.strftime('mmm dd yyyy HH:MM:ss')

    # Fetchers registry

    def register_fetcher(self, fetcher, rest_name):
        """ Register a children fetcher

        """
        self._fetchers_registry[rest_name] = fetcher

    def fetcher_for_rest_name(self, rest_name):
        """ Returns the children fetcher for the given rest name

            Args:
                rest_name (string): the children rest name

            Returns:
                list: Returns the corresponding fetcher

            Example:
                >>> print entity.fetcher_for_rest_name(NUSubEntity.rest_name)
                <NUSubEntitiesFetcher at yyyy>
        """
        if rest_name not in self._fetchers_registry:
            return None

        return self._fetchers_registry[rest_name]

    # Memory management

    def discard(self):
        """ Discard the current object and its children """

        if self._is_dirty:
            return

        self._is_dirty = True
        self.will_discard()

        bambou_logger.debug("[Bambou] Discarding object %s of type %s" % (self.id, self.rest_name))

        self.discard_all_fetchers()

        self._parent = None
        self._fetchers_registry = dict()

    def will_discard(self):
        """ """
        pass

    def discard_fetcher_for_rest_name(self, rest_name):
        """ Discard children with a given rest name

            Args:
                rest_name: the rest name

        """
        bambou_logger.debug("[Bambou] %s with id %s is discarding children %s" % (self.rest_name, self._id, rest_name))
        self._fetchers_registry[rest_name] = []

    def discard_all_fetchers(self):
        """ Discard current object children """

        for rest_name in self._fetchers_registry.keys():
            self.discard_fetcher_for_rest_name(rest_name)

    # Children management

    def add_child(self, child):
        """ Add a child """

        rest_name = child.rest_name
        children = self.fetcher_for_rest_name(rest_name)

        if children is None:
            raise InternalConsitencyError('Could not find fetcher with name %s while adding %s in parent %s' % (rest_name, child, self))

        if child not in children:
            children.append(child)

    def remove_child(self, child):
        """ Remove a child """

        rest_name = child.rest_name
        children = self.fetcher_for_rest_name(rest_name)

        target_child = None

        for local_child in children:
            if local_child.id == child.id:
                target_child = local_child
                break

        if target_child:
            children.remove(target_child)

    def update_child(self, child):
        """ Update child """

        rest_name = child.rest_name
        children = self.fetcher_for_rest_name(rest_name)

        index = None

        for local_child in children:
            if local_child.id == child.id:
                index = children.index(local_child)
                break

        if index:
            children[index] = child

    # Compression / Decompression

    def copy(self):
        """ Returns a copy of the current object

            Returns:
                A copy of the given object

            Example::
                >>> print entity.to_dict()
                <Entity object at 0x10b9d4910>

        """

        copy = self.__class__()

        for local_name, attribute in self._attributes.iteritems():
            value = getattr(self, local_name)

            if isinstance(value, NURESTObject):
                value = value.to_dict()

            if isinstance(value, list) and len(value) > 0 and isinstance(value[0], NURESTObject):
                tmp = list()
                for obj in value:
                    tmp.append(obj.to_dict())

                value = tmp

            setattr(copy, local_name, value)

        return copy

    def to_dict(self):
        """ Converts the current object into a Dictionary using all exposed ReST attributes.

            Returns:
                dict: the dictionary containing all the exposed ReST attributes and their values.

            Example::
                >>> print entity.to_dict()
                {"name": "my entity", "description": "Hello World", "ID": "xxxx-xxx-xxxx-xxx", ...}
        """

        dictionary = dict()

        for local_name, attribute in self._attributes.iteritems():
            remote_name = attribute.remote_name

            if hasattr(self, local_name):
                value = getattr(self, local_name)

                # Removed to resolve issue http://mvjira.mv.usa.alcatel.com/browse/VSD-5940 (12/15/2014)
                # if isinstance(value, bool):
                #     value = int(value)

                if isinstance(value, NURESTObject):
                    value = value.to_dict()

                if isinstance(value, list) and len(value) > 0 and isinstance(value[0], NURESTObject):
                    tmp = list()
                    for obj in value:
                        tmp.append(obj.to_dict())

                    value = tmp

                dictionary[remote_name] = value
            else:
                # print('Attribute %s could not be found for object %s' % (local_name, self))
                pass

        return dictionary

    def from_dict(self, dictionary):
        """ Sets all the exposed ReST attribues from the given dictionary

            Args:
                dictionary (dict): dictionnary containing the raw object attributes and their values.

            Example:
                >>> info = {"name": "my group", "private": False}
                >>> group = NUGroup()
                >>> group.from_dict(info)
                >>> print "name: %s - private: %s" % (group.name, group.private)
                "name: my group - private: False"
        """

        for remote_name, remote_value in dictionary.iteritems():
            # Check if a local attribute is exposed with the remote_name
            # if no attribute is exposed, return None
            local_name = next((name for name, attribute in self._attributes.iteritems() if attribute.remote_name == remote_name), None)

            if local_name:
                setattr(self, local_name, remote_value)
            else:
                # print('Attribute %s could not be added to object %s' % (remote_name, self))
                pass

    # HTTP Calls

    def delete(self, response_choice=1, async=False, callback=None):
        """ Delete object and call given callback in case of call.

            Args:
                response_choice (int): Automatically send a response choice when confirmation is needed
                async (bool): Boolean to make an asynchronous call. Default is False
                callback (function): Callback method that will be triggered in case of asynchronous call

            Example:
                >>> entity.delete() # will delete the enterprise from the server
        """
        return self._manage_child_object(nurest_object=self, method=HTTP_METHOD_DELETE, async=async, callback=callback, response_choice=response_choice)

    def save(self, response_choice=None, async=False, callback=None):
        """ Update object and call given callback in case of async call

            Args:
                async (bool): Boolean to make an asynchronous call. Default is False
                callback (function): Callback method that will be triggered in case of asynchronous call

            Example:
                >>> entity.name = "My Super Object"
                >>> entity.save() # will save the new name in the server
        """
        return self._manage_child_object(nurest_object=self, method=HTTP_METHOD_PUT, async=async, callback=callback, response_choice=response_choice)

    def fetch(self, async=False, callback=None):
        """ Fetch all information about the current object

            Args:
                async (bool): Boolean to make an asynchronous call. Default is False
                callback (function): Callback method that will be triggered in case of asynchronous call

            Returns:
                tuple: (current_fetcher, callee_parent, fetched_bjects, connection)

            Example:
                >>> entity = NUEntity(id="xxx-xxx-xxx-xxx")
                >>> entity.fetch() # will get the entity with id "xxx-xxx-xxx-xxx"
                >>> print entity.name
                "My Entity"
        """

        if self.id is None:
            raise InternalConsitencyError("Cannot fetch an object that does not have an ID")

        request = NURESTRequest(method=HTTP_METHOD_GET, url=self.get_resource_url())

        if async:
            return self.send_request(request=request, async=async, local_callback=self._did_retrieve, remote_callback=callback)
        else:
            connection = self.send_request(request=request)
            return self._did_retrieve(connection)

    # REST HTTP Calls

    def send_request(self, request, async=False, local_callback=None, remote_callback=None, user_info=None):
        """ Sends a request, calls the local callback, then the remote callback in case of async call

            Args:
                request: The request to send
                local_callback: local method that will be triggered in case of async call
                remote_callback: remote moethd that will be triggered in case of async call
                user_info: contains additionnal information to carry during the request

            Returns:
                Returns the object and connection (object, connection)
        """

        callbacks = dict()

        if local_callback:
            callbacks['local'] = local_callback

        if remote_callback:
            callbacks['remote'] = remote_callback

        connection = NURESTConnection(request=request, async=async, callback=self._did_receive_response, callbacks=callbacks)
        connection.user_info = user_info

        bambou_logger.info('Bambou Sending >>>>>>\n%s %s with following data:\n%s' % (request.method, request.url, json.dumps(request.data, indent=4)))

        return connection.start()

    def _manage_child_object(self, nurest_object, method=HTTP_METHOD_GET, async=False, callback=None, handler=None, response_choice=None, commit=False):
        """ Low level child management. Send given HTTP method with given nurest_object to given ressource of current object

            Args:
                nurest_object: the NURESTObject object to manage
                method: the HTTP method to use (GET, POST, PUT, DELETE)
                callback: the callback to call at the end
                handler: a custom handler to call when complete, before calling the callback
                commit: True to auto commit changes in the current object

            Returns:
                Returns the object and connection (object, connection)
        """
        url = None

        if method == HTTP_METHOD_POST:
            url = self.get_resource_url_for_child_type(nurest_object.__class__)
        else:
            url = self.get_resource_url()

        if response_choice is not None:
            url += '?responseChoice=%s' % response_choice

        request = NURESTRequest(method=method, url=url, data=nurest_object.to_dict())
        user_info = {'nurest_object': nurest_object, 'commit': commit}

        if not handler:
            handler = self._did_perform_standard_operation

        if async:

            return self.send_request(request=request, async=async, local_callback=handler, remote_callback=callback, user_info=user_info)
        else:
            connection = self.send_request(request=request, user_info=user_info)
            return handler(connection)


    # REST Operation handlers

    def _did_receive_response(self, connection):
        """ Receive a response from the connection """

        if connection.has_timeouted:
            bambou_logger.info("NURESTConnection has timeout.")
            return

        has_callbacks = connection.has_callbacks()
        should_post = not has_callbacks

        response_code = connection._response.status_code
        log_message = 'Bambou <<<<< Response [%s] for\n%s %s\n%s' % (response_code, connection._request.method, connection._request.url, json.dumps(connection._response.data, indent=4))
        level = logging.INFO

        if response_code >= 300:
            level = logging.WARNING

        bambou_logger.log(level, log_message)

        if  connection.handle_response_for_connection(should_post=should_post) and has_callbacks:
            callback = connection.callbacks['local']
            callback(connection)

    def _did_retrieve(self, connection):
        """ Callback called after fetching the object """

        response = connection.response

        try:
            self.from_dict(response.data[0])
        except:
            pass

        return self._did_perform_standard_operation(connection)

    def _did_perform_standard_operation(self, connection):
        """ Performs standard opertions """

        if connection.async:
            callback = connection.callbacks['remote']

            if connection.user_info and 'nurest_object' in connection.user_info:
                callback(connection.user_info['nurest_object'], connection)
            else:
                callback(self, connection)
        else:
            if connection.response.status_code >= 400 and BambouConfig._should_raise_bambou_http_error:
                raise BambouHTTPError(connection=connection)

            # Case with multiple objects like assignment
            if connection.user_info and 'nurest_objects' in connection.user_info:

                if connection.user_info['commit']:
                    for nurest_object in connection.user_info['nurest_objects']:
                        self.add_child(nurest_object)
                return (connection.user_info['nurest_objects'], connection)

            if connection.user_info and 'nurest_object' in connection.user_info:

                if connection.user_info['commit']:
                    self.add_child(connection.user_info['nurest_object'])

                return (connection.user_info['nurest_object'], connection)

            return (self, connection)

    # Advanced REST Operations

    def create_child(self, nurest_object, response_choice=None, async=False, callback=None, commit=True):
        """ Add given nurest_object to the current object

            For example, to add a child into a parent, you can call
            parent.create_child(nurest_object=child)

            Args:
                nurest_object (bambou.NURESTObject): the NURESTObject object to add
                response_choice (int): Automatically send a response choice when confirmation is needed
                async (bool): should the request be done asynchronously or not
                callback (function): callback containing the object and the connection

            Returns:
                Returns the object and connection (object, connection)

            Example:
                >>> entity = NUEntity(name="Super Entity")
                >>> parent_entity.create_child(entity) # the new entity as been created in the parent_entity
        """

        # if nurest_object.id:
        #     raise InternalConsitencyError("Cannot create a child that already has an ID: %s." % nurest_object)

        return self._manage_child_object(nurest_object=nurest_object,
                                         async=async,
                                         method=HTTP_METHOD_POST,
                                         callback=callback,
                                         handler=self._did_create_child,
                                         response_choice=response_choice,
                                         commit=commit)

    def instantiate_child(self, nurest_object, from_template, response_choice=None, async=False, callback=None, commit=True):
        """ Instantiate an nurest_object from a template object

            Args:
                nurest_object: the NURESTObject object to add
                from_template: the NURESTObject template object
                callback: callback containing the object and the connection

            Returns:
                Returns the object and connection (object, connection)

            Example:
                >>> parent_entity = NUParentEntity(id="xxxx-xxxx-xxx-xxxx") # create a NUParentEntity with an existing ID (or retrieve one)
                >>> other_entity_template = NUOtherEntityTemplate(id="yyyy-yyyy-yyyy-yyyy") # create a NUOtherEntityTemplate with an existing ID (or retrieve one)
                >>> other_entity_instance = NUOtherEntityInstance(name="my new instance") # create a new NUOtherEntityInstance to be intantiated from other_entity_template
                >>>
                >>> parent_entity.instantiate_child(other_entity_instance, other_entity_template) # instatiate the new domain in the server
        """

        # if nurest_object.id:
        #     raise InternalConsitencyError("Cannot instantiate a child that already has an ID: %s." % nurest_object)

        if not from_template.id:
            raise InternalConsitencyError("Cannot instantiate a child from a template with no ID: %s." % from_template)

        nurest_object.template_id = from_template.id
        return self._manage_child_object(nurest_object=nurest_object,
                                         async=async,
                                         method=HTTP_METHOD_POST,
                                         callback=callback,
                                         handler=self._did_create_child,
                                         response_choice=response_choice,
                                         commit=commit)

    def _did_create_child(self, connection):
        """ Callback called after adding a new child nurest_object """

        response = connection.response
        try:
            connection.user_info['nurest_object'].from_dict(response.data[0])
        except Exception:
            pass

        return self._did_perform_standard_operation(connection)

    def assign(self, objects, nurest_object_type, async=False, callback=None, commit=True):
        """ Reference a list of objects into the current resource

            Args:
                objects (list): list of NURESTObject to link
                nurest_object_type (type): Type of the object to link
                callback (function): Callback method that should be fired at the end

            Returns:
                Returns the current object and the connection (object, connection)

            Example:
                >>> entity.assign([entity1, entity2, entity3], NUEntity) # entity1, entity2 and entity3 are now part of the entity
        """

        ids = list()

        for nurest_object in objects:
            ids.append(nurest_object.id)

        url = self.get_resource_url_for_child_type(nurest_object_type)
        request = NURESTRequest(method=HTTP_METHOD_PUT, url=url, data=ids)
        user_info = {'nurest_objects': objects, 'commit': commit}

        if async:
            return self.send_request(request=request,
                              async=async,
                              local_callback=self._did_perform_standard_operation,
                              remote_callback=callback,
                              user_info=user_info)
        else:
            connection = self.send_request(request=request,
                                           user_info=user_info)

            return self._did_perform_standard_operation(connection)

    # Comparison

    def rest_equals(self, rest_object):
        """ Compare objects REST attributes

        """
        if not self.equals(rest_object):
            return False

        return self.to_dict() == rest_object.to_dict()

    def equals(self, rest_object):
        """ Compare with another object """

        if self._is_dirty:
            return False

        if rest_object is None:
            return False

        if not isinstance(rest_object, NURESTObject):
            raise TypeError('The object is not a NURESTObject %s' % rest_object)

        if self.rest_name != rest_object.rest_name:
            return False

        if self.id and rest_object.id:
            return self.id == rest_object.id

        if self.local_id and rest_object.local_id:
            return self.local_id == rest_object.local_id

        return False



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015, Alcatel-Lucent Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its contributors
#       may be used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from __future__ import print_function

try:
    import requests
    requests.packages.urllib3.disable_warnings()
except:
    pass

import logging

bambou_logger = logging.getLogger('bambou')
pushcenter_logger = logging.getLogger('pushcenter')

try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

bambou_logger.addHandler(NullHandler())

__all__ = ['NURESTBasicUser', 'NURESTConnection', 'NURESTModelController', 'NURESTFetcher', 'NURESTLoginController', 'NURESTObject', 'NURESTPushCenter', 'NURESTRequest', 'NURESTResponse', 'NURESTSession']

from bambou.nurest_session import NURESTSession
from bambou.nurest_user import NURESTBasicUser
from bambou.nurest_connection import NURESTConnection
from bambou.nurest_fetcher import NURESTFetcher
from bambou.nurest_login_controller import NURESTLoginController
from bambou.nurest_object import NURESTObject
from bambou.nurest_push_center import NURESTPushCenter
from bambou.nurest_request import NURESTRequest
from bambou.nurest_response import NURESTResponse
from bambou.nurest_modelcontroller import NURESTModelController
from bambou.config import BambouConfig



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015, Alcatel-Lucent Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its contributors
#       may be used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

try:
    import configparser
except ImportError:
    import ConfigParser as configparser


class BambouConfig(object):
    """ Bambou configuration

    """

    _should_raise_bambou_http_error = True
    _default_attribute_values_configuration_file_path = None
    _config_parser = None

    @classmethod
    def set_should_raise_bambou_http_error(cls, should_raise):
        """ Set if bambou should raise BambouHTTPError when
            a request fails

            Args:
                should_raise (bool): a boolean. Default is True.

        """
        cls._should_raise_bambou_http_error = should_raise

    @classmethod
    def set_default_values_config_file(cls, file_path):
        """ Set the name for an alternative default value configuration file

            Args:
                file_path (string): path of the config file

        """
        cls._default_attribute_values_configuration_file_path = file_path
        cls._read_config()

    @classmethod
    def _read_config(cls):
        """ Reads the configuration file if any
        """

        cls._config_parser = configparser.ConfigParser()
        cls._config_parser.read(cls._default_attribute_values_configuration_file_path)

    @classmethod
    def get_default_attribute_value(cls, object_class, property_name, attr_type=str):
        """ Gets the default value of a given property for a given object.

            These properties can be set in a config INI file looking like

            .. code-block:: ini

                [NUEntity]
                default_behavior = THIS
                speed = 1000

                [NUOtherEntity]
                attribute_name = a value

            This will be used when creating a :class:`bambou.NURESTObject` when no parameter or data is provided
        """

        if not cls._default_attribute_values_configuration_file_path:
            return None

        if not cls._config_parser:
            cls._read_config()

        class_name = object_class.__name__

        if not cls._config_parser.has_section(class_name):
            return None

        if not cls._config_parser.has_option(class_name, property_name):
            return None

        if attr_type in (int, long):
            return cls._config_parser.getint(class_name, property_name)
        elif attr_type is bool:
            return cls._config_parser.getboolean(class_name, property_name)
        else:
            return cls._config_parser.get(class_name, property_name)




# -*- coding: utf-8 -*-
#
# Copyright (c) 2015, Alcatel-Lucent Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its contributors
#       may be used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



class NURESTResponse(object):
    """ Response that will be received via the connection """

    def __init__(self, status_code, headers=None, data=None, reason=None, errors=None):
        """ Initializes a request """

        self._status_code = status_code
        self._data = data
        self._reason = reason
        self._headers = headers
        self.errors = dict()

    def __str__(self):
        """ Print request """

        return "[%s]\n%s" % (self.status_code, self.data)

    # Properties

    @property
    def status_code(self):
        """ Get method """

        return self._status_code

    @property
    def data(self):
        """ Get data """

        return self._data

    @property
    def reason(self):
        """ Get error reason """

        return self._reason

    @property
    def headers(self):
        """ Get headers """

        return self._headers



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015, Alcatel-Lucent Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its contributors
#       may be used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import requests
import threading
import uuid

from .nurest_response import NURESTResponse

from bambou import bambou_logger


HTTP_CODE_ZERO = 0
HTTP_CODE_SUCCESS = 200
HTTP_CODE_CREATED = 201
HTTP_CODE_EMPTY = 204
HTTP_CODE_MULTIPLE_CHOICES = 300
HTTP_CODE_BAD_REQUEST = 400
HTTP_CODE_UNAUTHORIZED = 401
HTTP_CODE_PERMISSION_DENIED = 403
HTTP_CODE_NOT_FOUND = 404
HTTP_CODE_METHOD_NOT_ALLOWED = 405
HTTP_CODE_CONNECTION_TIMEOUT = 408
HTTP_CODE_CONFLICT = 409
HTTP_CODE_PRECONDITION_FAILED = 412
HTTP_CODE_AUTHENTICATION_EXPIRED = 419
HTTP_CODE_INTERNAL_SERVER_ERROR = 500
HTTP_CODE_SERVICE_UNAVAILABLE = 503


HTTP_METHOD_HEAD = 'HEAD'
HTTP_METHOD_POST = 'POST'
HTTP_METHOD_GET = 'GET'
HTTP_METHOD_PUT = 'PUT'
HTTP_METHOD_DELETE = 'DELETE'


## TODO: remove this at some point. this is the SSL ugly patch
from requests.packages.urllib3.poolmanager import PoolManager
import ssl

## Monkey patch to use PROTOCOL_TLSv1 by default in requests
from functools import wraps
def sslwrap(func):
    @wraps(func)
    def bar(*args, **kw):
        kw['ssl_version'] = ssl.PROTOCOL_TLSv1
        return func(*args, **kw)
    return bar

PoolManager.__init__ = sslwrap(PoolManager.__init__)
## end of monkey patch


class NURESTConnection(object):
    """ Connection that enable HTTP requests """

    def __init__(self, request, async, callback=None, callbacks=dict(), user=None):
        """ Intializes a new connection for a given request

            NURESTConnection object is in charge of the HTTP call. It relies on request library

            Args:
                request: the NURESTRequest to send
                callback: the method that will be fired after sending
                callbacks: a dictionary of user callbacks. Should contains local and remote callbacks
        """

        self._uses_authentication = True
        self._has_timeouted = False
        #self._is_cancelled = False
        self._ignore_request_idle = False
        self._xhr_timeout = 3000
        self._response = None
        self._error_message = None
        self._transaction_id = uuid.uuid4().hex

        self._request = request
        self._async = async
        self._callback = callback
        self._callbacks = callbacks
        self._user_info = None
        self._object_last_action_timer = None
        self._user = user

    # Properties

    @property
    def callbacks(self):
        """ Get callbacks

            Returns:
                It returns an array containing user callbacks
        """

        return self._callbacks

    @property
    def request(self):
        """ Get request. Read-only property

            Returns:
                Returns the NURESTRequest object
        """

        return self._request

    @property
    def transaction_id(self):
        """ Get transaction ID. Read-only property

            Returns:
                Returns the transaction ID
        """

        return self._transaction_id

    @property
    def response(self):
        """ Get response

            Returns:
                It returns the NURESTResponse object of the request
        """

        return self._response

    @response.setter
    def response(self, response):
        """ Set response

            Args:
                response: the NURESTResponse object

        """

        self._response = response

    @property
    def user_info(self):
        """ Get user info

            Returns:
                It returns additionnal user information
        """

        return self._user_info

    @user_info.setter
    def user_info(self, info):
        """ Set user info

            Args:
                info: Information to carry
        """

        self._user_info = info

    @property
    def timeout(self):
        """ Get timeout

            Returns:
                It returns the timeout time in seconds. Default is 3000.
        """

        return self._xhr_timeout

    @timeout.setter
    def timeout(self, timeout):
        """ Set timeout

            Args:
                timeout: Number of seconds before timeout
        """

        self._xhr_timeout = timeout

    @property
    def ignore_request_idle(self):
        """ Get ignore request idle

            Returns:
                It returns a boolean. By default ignore request idle is set to False.
        """

        return self._ignore_request_idle

    @ignore_request_idle.setter
    def ignore_request_idle(self, ignore):
        """ Set ignore request idle

            Args:
                ignore: boolean to ignore request idle
        """

        self._ignore_request_idle = ignore

    @property
    def has_timeouted(self):
        """ Get has timouted

            Returns:
                Returns True if the request has timeout.
        """

        return self._has_timeouted

    @property
    def async(self):
        """ Get async

            Returns:
                Returns True if the request is asynchronous
        """

        return self._async

    # Methods

    def has_succeed(self):
        """ Check if the connection has succeed

            Returns:
                Returns True if connection has succeed.
                False otherwise.

        """
        status_code = self._response.status_code

        if status_code in [HTTP_CODE_ZERO, HTTP_CODE_SUCCESS, HTTP_CODE_CREATED, HTTP_CODE_EMPTY, HTTP_CODE_MULTIPLE_CHOICES]:
            return True

        if status_code in [HTTP_CODE_BAD_REQUEST, HTTP_CODE_UNAUTHORIZED, HTTP_CODE_PERMISSION_DENIED, HTTP_CODE_NOT_FOUND, HTTP_CODE_METHOD_NOT_ALLOWED, HTTP_CODE_CONNECTION_TIMEOUT, HTTP_CODE_CONFLICT, HTTP_CODE_PRECONDITION_FAILED, HTTP_CODE_INTERNAL_SERVER_ERROR, HTTP_CODE_SERVICE_UNAVAILABLE]:
            return False

        raise Exception('Unknown status code %s.', status_code)

    def has_callbacks(self):
        """  Check if the request has callbacks

            Returns:
                Returns YES if there is a local or remote callbacks
        """

        return len(self._callbacks) > 0

    def handle_response_for_connection(self, should_post=False):
        """ Check if the response succeed or not.

            In case of error, this method also print messages and set
            an array of errors in the response object.

            Returns:
                Returns True if the response has succeed, False otherwise
        """

        status_code = self._response.status_code
        data = self._response.data

        # TODO : Get errors in response data after bug fix : http://mvjira.mv.usa.alcatel.com/browse/VSD-2735
        if data and 'errors' in data:
            self._response.errors = data['errors']

        if status_code in [HTTP_CODE_SUCCESS, HTTP_CODE_CREATED, HTTP_CODE_EMPTY]:
            return True

        if status_code == HTTP_CODE_MULTIPLE_CHOICES:
            return False

        if status_code in [HTTP_CODE_PERMISSION_DENIED, HTTP_CODE_UNAUTHORIZED]:

            if not should_post:
                return True

            return False

        if status_code in [HTTP_CODE_CONFLICT, HTTP_CODE_NOT_FOUND, HTTP_CODE_BAD_REQUEST, HTTP_CODE_METHOD_NOT_ALLOWED, HTTP_CODE_PRECONDITION_FAILED, HTTP_CODE_SERVICE_UNAVAILABLE]:
            if not should_post:
                return True

            return False

        if status_code == HTTP_CODE_INTERNAL_SERVER_ERROR:

            return False

        if status_code == HTTP_CODE_ZERO:
            bambou_logger.error("NURESTConnection: Connection error with code 0. Sending NUNURESTConnectionFailureNotification notification and exiting.")
            return False

        bambou_logger.error("NURESTConnection: Report this error, because this should not happen: %s" % self._response)
        return False

    # HTTP Calls

    def _did_receive_response(self, response):
        """ Called when a response is received """

        try:
            data = response.json()

        except:
            data = None

        self._response = NURESTResponse(status_code=response.status_code, headers=response.headers, data=data, reason=response.reason)
        self._callback(self)

        return self

    def _did_timeout(self):
        """ Called when a resquest has timeout """

        bambou_logger.debug('Bambou %s on %s has timeout (timeout=%ss)..' % (self._request.method, self._request.url, self.timeout))
        self._has_timeouted = True

        if self.async:
            self._callback(self)
        else:
            return self

    def _make_request(self, session=None):
        """ Make a synchronous request """

        from .nurest_session import _NURESTSessionCurrentContext
        _NURESTSessionCurrentContext.session = session

        self._has_timeouted = False

        # Add specific headers
        controller = session.login_controller

        enterprise = controller.enterprise
        user_name = controller.user
        api_key = controller.api_key
        certificate = controller.certificate

        if self._user:
            enterprise = self._user.enterprise_name
            user_name = self._user.user_name
            api_key = self._user.api_key

        if self._uses_authentication:
            self._request.set_header('X-Nuage-Organization', enterprise)
            self._request.set_header('Authorization', controller.get_authentication_header(user_name, api_key))

        if controller.is_impersonating:
            self._request.set_header('X-Nuage-ProxyUser', controller.impersonation)

        headers = self._request.headers
        data = json.dumps(self._request.data)

        bambou_logger.debug('Bambou has been sent with user:%s within enterprise:%s (Key=%s)\nHeaders: %s' % (user_name, enterprise, api_key, headers))

        response = self.__make_request(method=self._request.method, url=self._request.url, data=data, headers=headers, certificate=certificate)

        retry_request = False

        if response.status_code == HTTP_CODE_MULTIPLE_CHOICES:
            self._request.url += '?responseChoice=1'
            bambou_logger.debug('Bambou got [%s] response. Trying to force response choice' % HTTP_CODE_MULTIPLE_CHOICES)
            retry_request = True

        elif response.status_code == HTTP_CODE_AUTHENTICATION_EXPIRED and _NURESTSessionCurrentContext.session:
            bambou_logger.debug('Bambou got [%s] response . Trying to reconnect your session that has expired' % HTTP_CODE_AUTHENTICATION_EXPIRED)
            _NURESTSessionCurrentContext.session.reset()
            _NURESTSessionCurrentContext.session.start()
            retry_request = True

        if retry_request:
            response = self.__make_request(method=self._request.method, url=self._request.url, data=data, headers=headers, certificate=certificate)

        return self._did_receive_response(response)

    def __make_request(self, method, url, data, headers, certificate):
        """ Encapsulate requests call
        """
        verify = False
        timeout = self.timeout

        try:  # TODO : Remove this ugly try/except after fixing Java issue: http://mvjira.mv.usa.alcatel.com/browse/VSD-546
            response = requests.request(  method=method,
                                          url=url,
                                          data=data,
                                          headers=headers,
                                          verify=verify,
                                          timeout=timeout,
                                          cert=certificate)
        except requests.exceptions.SSLError:
            try:
                response = requests.request(  method=method,
                                              url=url,
                                              data=data,
                                              headers=headers,
                                              verify=verify,
                                              timeout=timeout,
                                              cert=certificate)
            except requests.exceptions.Timeout:
                return self._did_timeout()

        except requests.exceptions.Timeout:
            return self._did_timeout()

        return response

    def start(self):
        """ Make an HTTP request with a specific method """

        # TODO : Use Timeout here and _ignore_request_idle
        from .nurest_session import NURESTSession
        session = NURESTSession.get_current_session()

        if self.async:
            thread = threading.Thread(target=self._make_request, kwargs={'session': session})
            thread.is_daemon = False
            thread.start()
            return self.transaction_id

        return self._make_request(session=session)

    def reset(self):
        """ Reset the connection

        """
        self._request = None
        self._response = None
        self._transaction_id = uuid.uuid4().hex



from types import ClassType, FunctionType
import sys, os
__all__ = [
    'decorate_class', 'metaclass_is_decorator', 'metaclass_for_bases',
    'frameinfo', 'decorate_assignment', 'decorate', 'struct', 'classy',
    'template_function', 'rewrap', 'cache_source', 'enclosing_frame',
    'synchronized',
]

def decorate(*decorators):
    """Use Python 2.4 decorators w/Python 2.3+

    Example::

        class Foo(object):
            decorate(classmethod)
            def something(cls,etc):
                \"""This is a classmethod\"""

    You can pass in more than one decorator, and they are applied in the same
    order that would be used for ``@`` decorators in Python 2.4.

    This function can be used to write decorator-using code that will work with
    both Python 2.3 and 2.4 (and up).
    """
    if len(decorators)>1:
        decorators = list(decorators)
        decorators.reverse()

    def callback(frame,k,v,old_locals):
        for d in decorators:
            v = d(v)
        return v
    return decorate_assignment(callback)

def enclosing_frame(frame=None, level=3):
    """Get an enclosing frame that skips DecoratorTools callback code"""
    frame = frame or sys._getframe(level)
    while frame.f_globals.get('__name__')==__name__: frame = frame.f_back
    return frame

def name_and_spec(func):
    from inspect import formatargspec, getargspec
    funcname = func.__name__
    if funcname=='<lambda>':
        funcname = "anonymous"
    args, varargs, kwargs, defaults = getargspec(func)
    return funcname, formatargspec(args, varargs, kwargs)[1:-1]


def qname(func):
    m = func.__module__
    return m and m+'.'+func.__name__ or func.__name__

class Bomb:
    def __str__(self):
        raise RuntimeError("template functions must return a static string!")
bomb = Bomb()

def getbody(func):
    from inspect import getargspec
    args, varargs, kwargs, defaults = getargspec(func)
    return func(*[bomb] * len(args))



















def apply_template(wrapper, func, *args, **kw):

    funcname, argspec  = name_and_spec(func)
    wrapname, wrapspec = name_and_spec(wrapper)

    body = wrapper.__doc__ or getbody(wrapper)
    if not body:
        raise RuntimeError(
            "Missing docstring or empty return value from"
            " %s(%s) - please switch the calling code from using docstrings"
            " to return values" % (wrapname, wrapspec)
        )

    body = body.replace('%','%%').replace('$args','%(argspec)s')
    body = """
def %(wrapname)s(%(wrapspec)s):
 def %(funcname)s(%(argspec)s): """+body+"""
 return %(funcname)s
"""
    body %= locals()

    filename = "<%s wrapping %s at 0x%08X>" % (qname(wrapper), qname(func), id(func))
    d ={}
    exec compile(body, filename, "exec") in func.func_globals, d

    f = d[wrapname](func, *args, **kw)
    cache_source(filename, body, f)

    f.func_defaults = func.func_defaults
    f.__doc__  = func.__doc__
    f.__dict__ = func.__dict__
    return f









def rewrap(func, wrapper):
    """Create a wrapper with the signature of `func` and a body of `wrapper`

    Example::

        def before_and_after(func):
            def decorated(*args, **kw):
                print "before"
                try:
                    return func(*args, **kw)
                finally:
                    print "after"
            return rewrap(func, decorated)

    The above function is a normal decorator, but when users run ``help()``
    or other documentation tools on the returned wrapper function, they will
    see a function with the original function's name, signature, module name,
    etc.

    This function is similar in use to the ``@template_function`` decorator,
    but rather than generating the entire decorator function in one calling
    layer, it simply generates an extra layer for signature compatibility.

    NOTE: the function returned from ``rewrap()`` will have the same attribute
    ``__dict__`` as the original function, so if you need to set any function
    attributes you should do so on the function returned from ``rewrap()``
    (or on the original function), and *not* on the wrapper you're passing in
    to ``rewrap()``.
    """
    def rewrap(__original, __decorated):
        return """return __decorated($args)"""
    return apply_template(rewrap, func, wrapper)









if sys.version<"2.5":
    # We'll need this for monkeypatching linecache
    def checkcache(filename=None):
        """Discard cache entries that are out of date.
        (This is not checked upon each call!)"""
        if filename is None:
            filenames = linecache.cache.keys()
        else:
            if filename in linecache.cache:
                filenames = [filename]
            else:
                return
        for filename in filenames:
            size, mtime, lines, fullname = linecache.cache[filename]
            if mtime is None:
                continue   # no-op for files loaded via a __loader__
            try:
                stat = os.stat(fullname)
            except os.error:
                del linecache.cache[filename]
                continue
            if size != stat.st_size or mtime != stat.st_mtime:
                del linecache.cache[filename]


def _cache_lines(filename, lines, owner=None):
    if owner is None:
        owner = filename
    else:
        from weakref import ref
        owner = ref(owner, lambda r: linecache and linecache.cache.__delitem__(filename))
    global linecache; import linecache
    if sys.version<"2.5" and linecache.checkcache.__module__!=__name__:
        linecache.checkcache = checkcache
    linecache.cache[filename] = 0, None, lines, owner

def cache_source(filename, source, owner=None):
    _cache_lines(filename, source.splitlines(True), owner)



def template_function(wrapper=None):
    """Decorator that uses its wrapped function's docstring as a template

    Example::

        def before_and_after(func):
            @template_function
            def wrap(__func, __message):
                '''
                print "before", __message
                try:
                    return __func($args)
                finally:
                    print "after", __message
                '''
            return wrap(func, "test")

    The above code will return individually-generated wrapper functions whose
    signature, defaults, ``__name__``, ``__module__``, and ``func_globals``
    match those of the wrapped functions.

    You can use define any arguments you wish in the wrapping function, as long
    as the first argument is the function to be wrapped, and the arguments are
    named so as not to conflict with the arguments of the function being
    wrapped.  (i.e., they should have relatively unique names.)

    Note that the function body will *not* have access to the globals of the
    calling module, as it is compiled with the globals of the *wrapped*
    function!  Thus, any non-builtin values that you need in the wrapper should
    be passed in as arguments to the template function.
    """
    if wrapper is None:
        return decorate_assignment(lambda f,k,v,o: template_function(v))
    return apply_template.__get__(wrapper)







def struct(*mixins, **kw):
    """Turn a function into a simple data structure class

    This decorator creates a tuple subclass with the same name and docstring as
    the decorated function.  The class will have read-only properties with the
    same names as the function's arguments, and the ``repr()`` of its instances
    will look like a call to the original function.  The function should return
    a tuple of values in the same order as its argument names, as it will be
    used by the class' constructor.  The function can perform validation, add
    defaults, and/or do type conversions on the values.

    If the function takes a ``*``, argument, it should flatten this argument
    into the result tuple, e.g.::

        @struct()
        def pair(first, *rest):
            return (first,) + rest

    The ``rest`` property of the resulting class will thus return a tuple for
    the ``*rest`` arguments, and the structure's ``repr()`` will reflect the
    way it was created.

    The ``struct()`` decorator takes optional mixin classes (as positional
    arguments), and dictionary entries (as keyword arguments).  The mixin
    classes will be placed before ``tuple`` in the resulting class' bases, and
    the dictionary entries will be placed in the class' dictionary.  These
    entries take precedence over any default entries (e.g. methods, properties,
    docstring, etc.) that are created by the ``struct()`` decorator.
    """
    def callback(frame, name, func, old_locals):

        def __new__(cls, *args, **kw):
            result = func(*args, **kw)
            if type(result) is tuple:
                return tuple.__new__(cls, (cls,)+result)
            else:
                return result

        def __repr__(self):
            return name+tuple.__repr__(self[1:])

        import inspect
        args, star, dstar, defaults = inspect.getargspec(func)

        d = dict(
            __new__ = __new__, __repr__ = __repr__, __doc__=func.__doc__,
            __module__ = func.__module__, __args__ = args, __star__ = star,
            __slots__ = [],
        )

        for p,a in enumerate(args):
            if isinstance(a,str):
                d[a] = property(lambda self, p=p+1: self[p])

        if star:
            d[star] = property(lambda self, p=len(args)+1: self[p:])

        d.update(kw)
        return type(name, mixins+(tuple,), d)

    return decorate_assignment(callback)





















def synchronized(func=None):
    """Create a method synchronized by first argument's ``__lock__`` attribute

    If the object has no ``__lock__`` attribute at run-time, the wrapper will
    attempt to add one by creating a ``threading.RLock`` and adding it to the
    object's ``__dict__``.  If ``threading`` isn't available, it will use a
    ``dummy_threading.RLock`` instead.  Neither will be imported unless the
    method is called on an object that doesn't have a ``__lock__``.

    This decorator can be used as a standard decorator (e.g. ``@synchronized``)
    or as a Python 2.3-compatible decorator by calling it with no arguments
    (e.g. ``[synchronized()]``).
    """
    if func is None:
        return decorate_assignment(lambda f,k,v,o: synchronized(v))

    from inspect import getargspec
    first_arg = getargspec(func)[0][0]

    def wrap(__func):
        return '''
        try:
            lock = $self.__lock__
        except AttributeError:
            try:
                from threading import RLock
            except ImportError:
                from dummy_threading import RLock
            lock = $self.__dict__.setdefault('__lock__',RLock())
        lock.acquire()
        try:
            return __func($args)
        finally:
            lock.release()'''.replace('$self', first_arg)

    return apply_template(wrap, func)





def frameinfo(frame):
    """Return (kind, module, locals, globals) tuple for a frame

    'kind' is one of "exec", "module", "class", "function call", or "unknown".
    """
    f_locals = frame.f_locals
    f_globals = frame.f_globals

    sameNamespace = f_locals is f_globals
    hasModule = '__module__' in f_locals
    hasName = '__name__' in f_globals
    sameName = hasModule and hasName
    sameName = sameName and f_globals['__name__']==f_locals['__module__']
    module = hasName and sys.modules.get(f_globals['__name__']) or None
    namespaceIsModule = module and module.__dict__ is f_globals

    if not namespaceIsModule:
        # some kind of funky exec
        kind = "exec"
        if hasModule and not sameNamespace:
            kind="class"
    elif sameNamespace and not hasModule:
        kind = "module"
    elif sameName and not sameNamespace:
        kind = "class"
    elif not sameNamespace:
        kind = "function call"
    else:
        # How can you have f_locals is f_globals, and have '__module__' set?
        # This is probably module-level code, but with a '__module__' variable.
        kind = "unknown"

    return kind,module,f_locals,f_globals








def decorate_class(decorator, depth=2, frame=None, allow_duplicates=False):

    """Set up `decorator` to be passed the containing class upon creation

    This function is designed to be called by a decorator factory function
    executed in a class suite.  The factory function supplies a decorator that
    it wishes to have executed when the containing class is created.  The
    decorator will be given one argument: the newly created containing class.
    The return value of the decorator will be used in place of the class, so
    the decorator should return the input class if it does not wish to replace
    it.

    The optional `depth` argument to this function determines the number of
    frames between this function and the targeted class suite.  `depth`
    defaults to 2, since this skips the caller's frame.  Thus, if you call this
    function from a function that is called directly in the class suite, the
    default will be correct, otherwise you will need to determine the correct
    depth value yourself.  Alternately, you can pass in a `frame` argument to
    explicitly indicate what frame is doing the class definition.

    This function works by installing a special class factory function in
    place of the ``__metaclass__`` of the containing class.  Therefore, only
    decorators *after* the last ``__metaclass__`` assignment in the containing
    class will be executed.  Thus, any classes using class decorators should
    declare their ``__metaclass__`` (if any) *before* specifying any class
    decorators, to ensure that all class decorators will be applied."""

    frame = enclosing_frame(frame, depth+1)
    kind, module, caller_locals, caller_globals = frameinfo(frame)

    if kind != "class":
        raise SyntaxError(
            "Class decorators may only be used inside a class statement"
        )
    elif not allow_duplicates and has_class_decorator(decorator, None, frame):
        return

    previousMetaclass = caller_locals.get('__metaclass__')
    defaultMetaclass  = caller_globals.get('__metaclass__', ClassType)


    def advise(name,bases,cdict):

        if '__metaclass__' in cdict:
            del cdict['__metaclass__']

        if previousMetaclass is None:
             if bases:
                 # find best metaclass or use global __metaclass__ if no bases
                 meta = metaclass_for_bases(bases)
             else:
                 meta = defaultMetaclass

        elif metaclass_is_decorator(previousMetaclass):
            # special case: we can't compute the "true" metaclass here,
            # so we need to invoke the previous metaclass and let it
            # figure it out for us (and apply its own advice in the process)
            meta = previousMetaclass

        else:
            meta = metaclass_for_bases(bases, previousMetaclass)

        newClass = meta(name,bases,cdict)

        # this lets the decorator replace the class completely, if it wants to
        return decorator(newClass)

    # introspection data only, not used by inner function
    # Note: these attributes cannot be renamed or it will break compatibility
    # with zope.interface and any other code that uses this decoration protocol
    advise.previousMetaclass = previousMetaclass
    advise.callback = decorator

    # install the advisor
    caller_locals['__metaclass__'] = advise


def metaclass_is_decorator(ob):
    """True if 'ob' is a class advisor function"""
    return isinstance(ob,FunctionType) and hasattr(ob,'previousMetaclass')


def iter_class_decorators(depth=2, frame=None):
    frame = enclosing_frame(frame, depth+1)
    m = frame.f_locals.get('__metaclass__')
    while metaclass_is_decorator(m):
        yield getattr(m, 'callback', None)
        m = m.previousMetaclass

def has_class_decorator(decorator, depth=2, frame=None):
    return decorator in iter_class_decorators(0, frame or sys._getframe(depth))
































def metaclass_for_bases(bases, explicit_mc=None):
    """Determine metaclass from 1+ bases and optional explicit __metaclass__"""

    meta = [getattr(b,'__class__',type(b)) for b in bases]

    if explicit_mc is not None:
        # The explicit metaclass needs to be verified for compatibility
        # as well, and allowed to resolve the incompatible bases, if any
        meta.append(explicit_mc)

    if len(meta)==1:
        # easy case
        return meta[0]

    classes = [c for c in meta if c is not ClassType]
    candidates = []

    for m in classes:
        for n in classes:
            if issubclass(n,m) and m is not n:
                break
        else:
            # m has no subclasses in 'classes'
            if m in candidates:
                candidates.remove(m)    # ensure that we're later in the list
            candidates.append(m)

    if not candidates:
        # they're all "classic" classes
        return ClassType

    elif len(candidates)>1:
        # We could auto-combine, but for now we won't...
        raise TypeError("Incompatible metatypes",bases)

    # Just one, return it
    return candidates[0]




def decorate_assignment(callback, depth=2, frame=None):
    """Invoke 'callback(frame,name,value,old_locals)' on next assign in 'frame'

    The frame monitored is determined by the 'depth' argument, which gets
    passed to 'sys._getframe()'.  When 'callback' is invoked, 'old_locals'
    contains a copy of the frame's local variables as they were before the
    assignment took place, allowing the callback to access the previous value
    of the assigned variable, if any.  The callback's return value will become
    the new value of the variable.  'name' is the name of the variable being
    created or modified, and 'value' is its value (the same as
    'frame.f_locals[name]').

    This function also returns a decorator function for forward-compatibility
    with Python 2.4 '@' syntax.  Note, however, that if the returned decorator
    is used with Python 2.4 '@' syntax, the callback 'name' argument may be
    'None' or incorrect, if the 'value' is not the original function (e.g.
    when multiple decorators are used).
    """
    frame = enclosing_frame(frame, depth+1)
    oldtrace = [frame.f_trace]
    old_locals = frame.f_locals.copy()

    def tracer(frm,event,arg):
        if event=='call':
            # We don't want to trace into any calls
            if oldtrace[0]:
                # ...but give the previous tracer a chance to, if it wants
                return oldtrace[0](frm,event,arg)
            else:
                return None

        try:
            if frm is frame and event !='exception':
                # Aha, time to check for an assignment...
                for k,v in frm.f_locals.items():
                    if k not in old_locals or old_locals[k] is not v:
                        break
                else:
                    # No luck, keep tracing
                    return tracer

                # Got it, fire the callback, then get the heck outta here...
                frm.f_locals[k] = callback(frm,k,v,old_locals)

        finally:
            # Give the previous tracer a chance to run before we return
            if oldtrace[0]:
                # And allow it to replace our idea of the "previous" tracer
                oldtrace[0] = oldtrace[0](frm,event,arg)

        uninstall()
        return oldtrace[0]

    def uninstall():
        # Unlink ourselves from the trace chain.
        frame.f_trace = oldtrace[0]
        sys.settrace(oldtrace[0])

    # Install the trace function
    frame.f_trace = tracer
    sys.settrace(tracer)

    def do_decorate(f):
        # Python 2.4 '@' compatibility; call the callback
        uninstall()
        frame = sys._getframe(1)
        return callback(
            frame, getattr(f,'__name__',None), f, frame.f_locals
        )

    return do_decorate











def super_next(cls, attr):
    for c in cls.__mro__:
        if attr in c.__dict__:
            yield getattr(c, attr).im_func

# Python 2.6 and above mix ABCMeta into various random places  :-(
try:
    from abc import ABCMeta as base
except ImportError:
    base = type

class classy_class(base):
    """Metaclass that delegates selected operations back to the class"""

    def __new__(meta, name, bases, cdict):
        cls = super(classy_class, meta).__new__(meta, name, bases, cdict)
        supr = super_next(cls, '__class_new__').next
        return supr()(meta, name, bases, cdict, supr)

    def __init__(cls, name, bases, cdict):
        supr = super_next(cls, '__class_init__').next
        return supr()(cls, name, bases, cdict, supr)

    def __call__(cls, *args, **kw):
        return cls.__class_call__.im_func(cls, *args, **kw)

    if base is not type:
        # Our instances do not support ABC-ness
        def register(*args): raise NotImplementedError
        __instancecheck__ = type.__dict__['__instancecheck__']
        __subclasscheck__ = type.__dict__['__subclasscheck__']











class classy(object):
    """Base class for classes that want to be their own metaclass"""
    __metaclass__ = classy_class
    __slots__ = ()

    def __class_new__(meta, name, bases, cdict, supr):
        return type.__new__(meta, name, bases, cdict)

    def __class_init__(cls, name, bases, cdict, supr):
        return type.__init__(cls, name, bases, cdict)

    def __class_call__(cls, *args, **kw):
        return type.__call__(cls, *args, **kw)
    __class_call__ = classmethod(__class_call__)































# -*- coding: utf-8 -*-

# Copyright 2014 Alcatel-Lucent USA Inc.

from .nurest_login_controller import NURESTLoginController
from .nurest_push_center import NURESTPushCenter
from bambou.contextual import context
from bambou import bambou_logger
from contextlib import contextmanager
from opcode import opname
import inspect


class NURESTSession(object):

    """ Authenticated sessions used for sending ReST calls

        The session holds the credential information for a particular
        user. It is used by :class:`bambou.NURESTConnection` to get
        the authentication information needed to send the request.

        NURESTSession supports the `with` statement.

        Note:
            NURESTSession *must* be subclassed, and the subclass *must* implement :class:`bambou.NURESTSession.create_rest_user`

        Example:
            >>> mainsession =  NUMySession(username="csproot", password="csproot", enterprise="csp", api_url="https://vsd:8443", version="3.2")
            >>> othersession = NUMySession(username="user", password="password", enterprise="ent", api_url="https://vsd:8443", version="3.2")
            >>>
            >>> mainsession.start()
            >>> mainsession.user.entities.get()
            [<NUEntity at 1>, <NUEntity at 2>, <NUEntity at 3>]
            >>>
            >>> with othersession.start() as session:
            >>>     session.user.entities.get()
            [<NUEntity at 2>]
    """

    def __init__(self, username, password, enterprise, api_url, version, certificate=None):
        """ Initializes a new sesssion

            Args:
                username (string): the username
                password (string): the password
                enterprise (string): the enterprise
                api_url (string): the url to the api
                version (string): the version of the api to target

            Example:
                >>> mainsession =  NUMySession(username="csproot", password="csproot", enterprise="csp", api_url="https://vsd:8443", version="3.2")

        """
        self._user = None
        self._login_controller = NURESTLoginController()
        self._login_controller.user = username
        self._login_controller.password = password
        self._login_controller.certificate = certificate
        self._login_controller.user_name = username
        self._login_controller.enterprise = enterprise
        self._login_controller.url = '%s/nuage/api/v%s' % (api_url, str(version).replace('.', '_'))

        self._push_center = NURESTPushCenter()
        self._push_center.url = self._login_controller.url

    # Class Methods

    @classmethod
    def get_current_session(cls):
        """
            Get the current session

            Returns:
                (bambou.NURESTSession): the current session

        """
        return _NURESTSessionCurrentContext.session

    # Properties

    @property
    def push_center(self):
        """
            Returns the :class:`bambou.NURESTPushCenter` of the current session

            Note:
                Use this method to start and stop receiving push notifications
        """
        return self._push_center

    @property
    def login_controller(self):
        """
            Returns the :class:`bambou.NURESTLoginController` of the current session

            Note:
                You should not need to use this method. It's used automatically when needed
        """
        return self._login_controller

    @property
    def user(self):
        """
            Returns the user of the session

            Returns:
                (bambou.NURESTBasicUser): the REST user
        """
        return self._user

    @property
    def is_impersonating(self):
        """ Returns True if the session is currently impersonating
            a user

            Returns:
                (bool): a boolean that indicate if the session is impersonating a user

        """
        return self._login_controller.is_impersonating

    # Methods

    def create_rest_user(self):
        """
            Create a :class:`bambou.NURESTBasicUser`.

            This method *MUST* be overriden by subclasses in order to provide
            a valid :class:`bambou.NURESTBasicUser`.

            Returns:
                A instance of a subclass of :class:`bambou.NURESTBasicUser`
        """
        raise NotImplementedError('%s must define method def create_rest_user(self).' % self)

    def _authenticate(self):

        if self._user is None:
            self._user = self.create_rest_user()
            self._user.fetch()

        self.login_controller.api_key = self._user.api_key
        bambou_logger.debug("[NURESTSession] Started session with username %s in enterprise %s" % (self.login_controller._user, self.login_controller.enterprise))

    def _in_with_statement(self, frame):
        """
        Well, you'll have to trust me on this one.
        """
        return opname[ord(frame.f_code.co_code[frame.f_lasti + 3])] is "SETUP_WITH"


    def start(self):
        """
            Starts the session.

            Starting the session will actually get the API key of the current user
        """

        try:
            frame = inspect.stack()[1][0]
        except IndexError:
            _NURESTSessionCurrentContext.session = self
            self._authenticate()
            return self

        if self._in_with_statement(frame):
            return _NURESTSessionContext.new(self)
        else:
            _NURESTSessionCurrentContext.session = self
            self._authenticate()
            return self

    def reset(self):
        """
            Resets the session.

            Resetting the session will flush the API stored API key. Any additional calls will require to call start, and a
            /me request will be reissued.
        """

        self._user = None
        self.login_controller.api_key = None

    def impersonate(self, username, enterprise):
        """
            Change the session to impersonate a user within an enterprise

            Args:
                username (string): name of the user to impersonate
                enterprise (string): name of the enterprise
        """
        self._login_controller.impersonate(username, enterprise)

    def stop_impersonate(self):
        """
            Stop impersonating a user

        """
        self._login_controller.stop_impersonate()

    def equals(self, session):
        """ Verify if the current session equals the given parameter

            Notes:
                Verification is based on username, enterprise, api_url and its version.

            Args:
                session(bambou.NURESTSession): the session to compare with

            Returns:
                (bool): True if session are equal
        """
        return self.login_controller.equals(session.login_controller)

    def is_current_session(self):
        """ Verify if the session is the current.

            Returns:
                (bool): True if the session is the current
        """
        current_session = NURESTSession.get_current_session()
        return current_session and self.equals(current_session)


class _NURESTSessionContext (object):

    session = None

    @classmethod
    @contextmanager
    def new(self, session):
        with _NURESTSessionCurrentContext.new() as context:
            context.session = session
            session._authenticate()
            yield session


class _NURESTSessionCurrentContext (context.Service):

    session = None



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015, Alcatel-Lucent Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its contributors
#       may be used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
from setuptools  import setup

version = os.environ.get("BAMBOU_VERSION", "1.0.0")

setup(
    name='bambou',
    version=version,
    author='Christophe Serafin, Antoine Mercadal',
    author_email='christophe.serafin@nuagenetworks.net, antoine@nuagenetworks.net',
    packages=['bambou', 'bambou.utils', 'bambou.contextual'],
    description='REST Library for Nuage Networks',
    long_description=open('README.md').read(),
    install_requires=[line for line in open('requirements.txt')],
    license='BSD-3',
    url='https://github.com/nuagenetworks',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: Utilities",
        "License :: OSI Approved :: BSD License",
        "Environment :: Console",
        "Intended Audience :: Developers"
    ]
)



# -*- coding: utf-8 -*-

#
# Copyright (c) 2015, Alcatel-Lucent Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its contributors
#       may be used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from .exceptions import BambouHTTPError, InternalConsitencyError
from .nurest_request import NURESTRequest
from .nurest_connection import HTTP_METHOD_GET, HTTP_METHOD_HEAD

from bambou.config import BambouConfig


class NURESTFetcher(list):
    """ Object fetcher for childrens

        This object is intended to fetch a specific type of object declared
        by `managed_class` method
    """

    PAGE_SIZE = 50

    def __init__(self):
        """ Initliazes the fetcher """

        super(NURESTFetcher, self).__init__()

        self.current_connection = None
        self.current_ordered_by = ''
        self.current_page = 0
        self.current_total_count = 0
        self.query_string = None
        self._parent_object = None

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__, super(NURESTFetcher, self).__repr__())

    def __contains__(self, nurest_object):
        """ Verify if the fetcher contains the given NURESTObject

            Args:
                nurest_object (bambou.NURESTObject): the NURESTObject object to verify

            Returns:
                Returns True if the object has been found. False otherwise

        """
        for obj in self:
            if obj.equals(nurest_object):
                return True

        return False

    def index(self, nurest_object):
        """ Get index of the given item
            Args:
                nurest_object (bambou.NURESTObject): the NURESTObject object to verify

            Returns:
                Returns the position of the object.

            Raises:
                Raise a ValueError exception if object is not present
        """
        for index, obj in enumerate(self):
            if obj.equals(nurest_object):
                return index

        raise ValueError("%s is  not in %s" % (nurest_object, self))

    # Properties

    @property
    def parent_object(self):
        """ Get served object

            The fetcher will fill the served object with fetched objects

            Returns:
                Returns the object it is serving.
        """

        return self._parent_object

    @parent_object.setter
    def parent_object(self, parent_object):
        """ Set the served object

            Args:
                parent_object: the object to serve
        """

        self._parent_object = parent_object

    @property
    def transaction_id(self):
        """ Get the transaction ID of the current connection

            Returns:
                the transaction id
        """
        if self.current_connection is None:
            raise InternalConsitencyError("%s could not find a current transaction." % self)

        return self.current_connection.transaction_id

    # Methods

    @classmethod
    def managed_class(cls):
        """ Defines the object type that is managed by this fetcher.

            This method has to be overrided to indicate which object the fetcher
            is allowed to fetch.

            Returns:
                Returns a type of NURESTObject

            Raises:
                NotImplementedError: if this method is not overriden

        """

        raise NotImplementedError('%s has no managed class. Implements managed_class method first.' % cls)

    @classmethod
    def managed_object_rest_name(cls):
        """ Remote name of the managed object

            Returns:
                Returns a string representing the resource name of the object
                that has to be fetched.
        """

        return cls.managed_class().rest_name

    @classmethod
    def fetcher_with_object(cls, parent_object):
        """ Register the fetcher for a served object.

            This method will fill the fetcher with `managed_class` instances

            Args:
                parent_object: the instance of the parent object to serve

            Returns:
                It returns the fetcher instance.
        """

        fetcher = cls()
        fetcher.parent_object = parent_object

        rest_name = cls.managed_object_rest_name()
        parent_object.register_fetcher(fetcher, rest_name)

        return fetcher

    def flush(self):
        """ Removes all fetched objects

            It will clear attribute of the served object
        """
        self.current_connection = None
        del self[:]

    def new(self):
        """ Create an instance of the managed class

            Returns:
                Returns an instance of managed_class
        """

        managed_class = self.managed_class()
        return managed_class()

    def _prepare_headers(self, request, filter=None, order_by=None, group_by=[], page=None, page_size=None):
        """ Prepare headers for the given request

            Args:
                request: the NURESTRequest to send
                filter: string
                order_by: string
                group_by: list of names
                page: int
                page_size: int
        """

        if filter:
            request.set_header('X-Nuage-Filter', filter)

        if order_by:
            request.set_header('X-Nuage-OrderBy', order_by)

        if page:
            request.set_header('X-Nuage-Page', page)

        if page_size:
            request.set_header('X-Nuage-PageSize', page_size)

        if len(group_by) > 0:
            header = ", ".join(group_by)
            request.set_header('X-Nuage-GroupBy', 'true')
            request.set_header('X-Nuage-Attributes', header)

    def _prepare_url(self):
        """ Prepare url for request """

        url = self.parent_object.get_resource_url_for_child_type(self.__class__.managed_class())

        if self.query_string:
            url += "?%s" % self.query_string

        return url

    def fetch(self, filter=None, order_by=None, group_by=[], page=None, page_size=None, commit=True, async=False, callback=None):
        """ Fetch objects according to given filter and page.

            Note:
                This method fetches all managed class objects and store them
                in local_name of the served object. which means that the parent
                object will hold them in a list. You can prevent this behavior
                by setting commit to False. In that case, the fetched children
                won't be added in the parent object cache.

            Args:
                filter (string): string that represents a predicate filter
                order_by (string): string that represents an order by clause
                group_by (string): list of names for grouping
                page (int): number of the page to load
                page_size (int): number of results per page
                commit (bool): boolean to update current object
                callback (function): Callback that should be called in case of a async request

            Returns:
                tuple: Returns a tuple of information (fetcher, served object, fetched objects, connection)

            Example:
                >>> entity.children.fetch()
                (<NUChildrenFetcher at aaaa>, <NUEntity at bbbb>, [<NUChildren at ccc>, <NUChildren at ddd>], <NURESTConnection at zzz>)
        """

        request = NURESTRequest(method=HTTP_METHOD_GET, url=self._prepare_url())

        self._prepare_headers(request=request, filter=filter, order_by=order_by, group_by=group_by, page=page, page_size=page_size)

        if async:
            return self.parent_object.send_request(request=request, async=async, local_callback=self._did_fetch, remote_callback=callback, user_info={'commit': commit})

        connection = self.parent_object.send_request(request=request, user_info={'commit': commit})
        return self._did_fetch(connection=connection)

    def _did_fetch(self, connection):
        """ Fetching objects has been done """

        self.current_connection = connection
        response = connection.response
        should_commit = 'commit' not in connection.user_info or connection.user_info['commit']

        if connection.response.status_code >= 400 and BambouConfig._should_raise_bambou_http_error:
            raise BambouHTTPError(connection=connection)

        if response.status_code != 200:

            if should_commit:
                self.current_total_count = 0
                self.current_page = 0
                self.current_ordered_by = ''

            return self._send_content(content=None, connection=connection)

        results = response.data
        fetched_objects = list()

        if should_commit:
            if 'X-Nuage-Count' in response.headers and response.headers['X-Nuage-Count']:
                self.current_total_count = int(response.headers['X-Nuage-Count'])

            if 'X-Nuage-Page' in response.headers and response.headers['X-Nuage-Page']:
                self.current_page = int(response.headers['X-Nuage-Page'])

            if 'X-Nuage-OrderBy' in response.headers and response.headers['X-Nuage-OrderBy']:
                self.current_ordered_by = response.headers['X-Nuage-OrderBy']

        if results:
            for result in results:

                nurest_object = self.new()
                nurest_object.from_dict(result)
                nurest_object.parent = self.parent_object

                fetched_objects.append(nurest_object)

                if not should_commit:
                    continue

                if nurest_object in self:
                    idx = self.index(nurest_object)
                    current_object = self[idx]
                    current_object.from_dict(nurest_object.to_dict())
                else:
                    self.append(nurest_object)

        return self._send_content(content=fetched_objects, connection=connection)

    def get(self, filter=None, order_by=None, group_by=[], page=None, page_size=None, commit=True, async=False, callback=None):
        """ Fetch object and directly return them

            Note:
                `get` won't put the fetched objects in the parent's children list.
                You cannot override this behavior. If you want to commit them in the parent
                you can use :method:vsdk.NURESTFetcher.fetch or manually add the list with
                :method:vsdk.NURESTObject.add_child

            Args:
                filter (string): string that represents a predicate filter
                order_by (string): string that represents an order by clause
                group_by (string): list of names for grouping
                page (int): number of the page to load
                page_size (int): number of results per page
                commit (bool): boolean to update current object
                callback (function): Callback that should be called in case of a async request

            Returns:
                list: list of vsdk.NURESTObject if any

            Example:
                >>> print entity.children.get()
                [<NUChildren at xxx>, <NUChildren at yyyy>, <NUChildren at zzz>]
        """
        return self.fetch(filter=filter, order_by=order_by, group_by=group_by, page=page, page_size=page_size, commit=False)[2]

    def get_first(self, filter=None, order_by=None, group_by=[], page=None, page_size=None, commit=True, async=False, callback=None):
        """ Fetch object and directly return the first one

            Note:
                `get_first` won't put the fetched object in the parent's children list.
                You cannot override this behavior. If you want to commit it in the parent
                you can use :method:vsdk.NURESTFetcher.fetch or manually add it with
                :method:vsdk.NURESTObject.add_child

            Args:
                filter (string): string that represents a predicate filter
                order_by (string): string that represents an order by clause
                group_by (string): list of names for grouping
                page (int): number of the page to load
                page_size (int): number of results per page
                commit (bool): boolean to update current object
                callback (function): Callback that should be called in case of a async request

            Returns:
                vsdk.NURESTObject: the first object if any, or None

            Example:
                >>> print entity.children.get_first(filter="name == 'My Entity'")
                <NUChildren at xxx>
        """
        objects = self.get(filter=filter, order_by=order_by, group_by=group_by, page=0, page_size=1, commit=False)
        return objects[0] if len(objects) else None

    def count(self, filter=None, order_by=None, group_by=[], page=None, page_size=None, async=False, callback=None):
        """ Get the total count of objects that can be fetched according to filter

            This method can be asynchronous and trigger the callback method
            when result is ready.

            Args:
                filter (string): string that represents a predicate fitler (eg. name == 'x')
                order_by (string): string that represents an order by clause
                group_by (string): list of names for grouping
                page (int): number of the page to load
                page_size (int): number of results per page
                callback (function): Method that will be triggered asynchronously

            Returns:
                Returns a transaction ID when asynchronous call is made.
                Otherwise it will return a tuple of information containing
                (fetcher, served object, count of fetched objects)
        """

        request = NURESTRequest(method=HTTP_METHOD_HEAD, url=self._prepare_url())

        self._prepare_headers(request=request, filter=filter, order_by=order_by, group_by=group_by, page=page, page_size=page_size)

        if async:
            return self.parent_object.send_request(request=request, async=async, local_callback=self._did_count, remote_callback=callback)

        else:
            connection = self.parent_object.send_request(request=request)
            return self._did_count(connection)

    def get_count(self, filter=None, order_by=None, group_by=[], page=None, page_size=None):
        """ Get the total count of objects that can be fetched according to filter

            Args:
                filter (string): string that represents a predicate fitler (eg. name == 'x')
                order_by (string): string that represents an order by clause
                group_by (string): list of names for grouping
                page (int): number of the page to load
                page_size (int): number of results per page

            Returns:
                Returns the number of objects found

        """
        return self.count(filter=filter, order_by=order_by, group_by=group_by, page=page, page_size=page_size, async=False)[2]

    def _did_count(self, connection):
        """ Called when count if finished """

        self.current_connection = connection
        response = connection.response
        count = 0
        callback = None

        if 'X-Nuage-Count' in response.headers:
            count = int(response.headers['X-Nuage-Count'])

        if 'remote' in connection.callbacks:
            callback = connection.callbacks['remote']

        if connection.async:
            if callback:
                callback(self, self.parent_object, count)

                self.current_connection.reset()
                self.current_connection = None
        else:

            if connection.response.status_code >= 400 and BambouConfig._should_raise_bambou_http_error:
                raise BambouHTTPError(connection=connection)

            return (self, self.parent_object, count)

    def _send_content(self, content, connection):
        """ Send a content array from the connection """

        if connection:

            if connection.async:
                callback = connection.callbacks['remote']

                if callback:
                    callback(self, self.parent_object, content)

                    self.current_connection.reset()
                    self.current_connection = None
            else:
                return (self, self.parent_object, content)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2015, Alcatel-Lucent Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its contributors
#       may be used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import json
import threading

from time import time

from .nurest_connection import NURESTConnection
from .nurest_request import NURESTRequest

from bambou import pushcenter_logger

class StoppableThread(threading.Thread):
    """Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition."""

    def __init__(self, *args, **kwargs):
        super(StoppableThread, self).__init__(*args, **kwargs)
        self._stop = threading.Event()

    def stop(self):
        self._stop.set()

    def stopped(self):
        return self._stop.isSet()


class NURESTPushCenter(object):
    """ Push center wait for push notifications.

        It has to listen a specific URL.

        Every time a notification is send, it will automatically get it
        and store it into get_last_events method.
    """

    def __init__(self):
        """ Initialize push center """

        self._url = None
        self._is_running = False
        self._current_connection = None
        self._last_events = list()
        self.nb_events_received = 0
        self.nb_push_received = 0
        self._thread = None
        self._user = None
        self._start_time = None
        self._timeout = None
        self._delegate_methods = list()

    # Properties

    @property
    def url(self):
        """ Get url """

        return self._url

    @url.setter
    def url(self, url):
        """ Set url """

        self._url = url

    @property
    def is_running(self):
        """ Get is_running """

        return self._is_running

    # Control Methods

    def start(self, timeout=None, user=None):
        """ Starts listening to events.

            Args:
                timeout (int): number of seconds before timeout. Used for testing purpose only.
                user (bambou.NURESTBasicUser): NURESTBasicUser object that is listening. Used for testing purpose only.
        """

        if self._is_running:
            return

        if timeout:
            self._timeout = timeout
            self._start_time = int(time())

        pushcenter_logger.debug("[NURESTPushCenter] Starting push center on url %s ..." % self.url)
        self._is_running = True
        self._user = user

        from .nurest_session import NURESTSession
        current_session = NURESTSession.get_current_session()
        args_session = {'session': current_session}

        self._thread = StoppableThread(target=self._listen, name='push-center', kwargs=args_session)
        self._thread.daemon = True
        self._thread.start()

    def stop(self):
        """ Stops listening for events. """

        if not self._is_running:
            return

        pushcenter_logger.debug("[NURESTPushCenter] Stopping...")

        self._thread.stop()
        self._thread.join()

        self._is_running = False
        self._current_connection = None
        self._start_time = None
        self._timeout = None

    def wait_until_exit(self):
        """ Wait until thread exit

            Used for testing purpose only
        """

        if self._timeout is None:
            raise Exception("Thread will never exit. Use stop or specify timeout when starting it!")

        self._thread.join()
        self.stop()

    # Events

    def get_last_events(self):
        """ Retrieve events that has been

            Returns:
                Returns a list of events and flush existing events.
        """

        events = self._last_events
        self._last_events = list()
        return events

    # Private methods

    def _did_receive_event(self, connection):
        """ Receive an event from connection """

        if not self._is_running:
            return

        if connection.has_timeouted:
            return

        response = connection.response

        if response.status_code != 200:
            pushcenter_logger.error("[NURESTPushCenter]: Connection failure [%s] %s" % (response.status_code, response.errors))

        else:
            data = response.data

            if len(self._delegate_methods) > 0:
                for m in self._delegate_methods:
                    try:
                        m(data)
                    except Exception as exc:
                        pushcenter_logger.error("[NURESTPushCenter] Delegate method %s failed:\n%s" % (m, exc))
            elif data:
                events = data['events']
                self.nb_events_received += len(events)
                self.nb_push_received += 1

                pushcenter_logger.info("[NURESTPushCenter] Received Push #%s (total=%s, latest=%s)\n%s" % (self.nb_push_received, self.nb_events_received, len(events), json.dumps(events, indent=4)))
                self._last_events.extend(events)

        if self._is_running:
            uuid = None
            if data and 'uuid' in data:
                uuid = data['uuid']

            self._listen(uuid)

    def _listen(self, uuid=None, session=None):
        """ Listen a connection uuid """
        if session:
            from .nurest_session import _NURESTSessionCurrentContext
            _NURESTSessionCurrentContext.session = session

        if self.url is None:
            raise Exception("NURESTPushCenter needs to have a valid URL. please use setURL: before starting it.")

        events_url = "%s/events" % self.url
        if uuid:
            events_url = "%s?uuid=%s" % (events_url, uuid)

        request = NURESTRequest(method='GET', url=events_url)

        # Force async to False so the push center will have only 1 thread running
        connection = NURESTConnection(request=request, async=True, callback=self._did_receive_event, user=self._user)

        if self._timeout:
            if int(time()) - self._start_time >= self._timeout:
                pushcenter_logger.debug("[NURESTPushCenter] Timeout (timeout=%ss)." % self._timeout)
                return

            else:
                connection.timeout = self._timeout

        pushcenter_logger.info('Bambou Sending >>>>>>\n%s %s' % (request.method, request.url))

        #connection.ignore_request_idle = True
        connection.start()

    def add_delegate(self, callback):
        """ Registers a new delegate callback

            The prototype should be function(data), where data will be the decoded json push

            Args:
                callback (function): method to trigger when push center receives events
        """

        if callback in self._delegate_methods:
            return

        self._delegate_methods.append(callback)

    def remove_delegate(self, callback):
        """ Unregisters a registered delegate function or a method.

            Args:
                callback(function): method to trigger when push center receives events
        """

        if not callback in self._delegate_methods:
            return

        self._delegate_methods.remove(callback)
